#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame import plugins
from libqtopensesame.misc.config import cfg
from libqtopensesame.widgets.base_widget import base_widget
try:
	from qtpy.QtCore import QPyNullVariant
except ImportError:
	QPyNullVariant = type(None)

class plugin_widget(base_widget):

	"""
	desc:
		A widget for a single plugin.
	"""

	def __init__(self, plugin, main_window):

		"""
		desc:
			Constructor.

		arguments:
			plugin:			The name of the plugin.
			main_window:	The main-window object.
		"""

		super(plugin_widget, self).__init__(main_window,
			ui=u'extensions.plugin_manager.plugin')
		self.plugin = plugin
		self.info = plugins.plugin_properties(plugin)
		self.ui.label_name.setText(plugin)
		self.ui.label_folder.setText(self.info[u'plugin_folder'])
		if u'description' in self.info:
			self.ui.label_description.setText(safe_decode(
				self.info[u'description']))
		if u'author' in self.info:
			self.ui.label_author.setText(safe_decode(self.info[u'author']))
		if u'version' in self.info:
			self.ui.label_version.setText(safe_decode(self.info[u'version']))
		if u'url' in self.info:
			self.ui.label_url.setText(safe_decode(self.info[u'url']))
		if u'startup_time' in self.info:
			self.ui.label_startup_time.setText(
				u'%.4f s' % self.info[u'startup_time']
			)
		self.ui.checkbox_enable.setChecked(self.is_enabled())
		self.ui.checkbox_enable.clicked.connect(self.toggle)

	def is_enabled(self):

		"""
		returns:
			True if the plug-in is enabled, False otherwise.
		"""

		cfg_var = u'disabled_%s' % self.info[u'type']
		# On Mac (at least) cfg['disabled_plugins'] will return a QPyNullVariant type
		# when the disabled plugin list is empty. the 'in' operator is not allowed on
		# this datatype. It is reasonable to assume that the plugin is not disabled if
		# the disabled list is empty and thus a QPyNullVariant.
		if isinstance(cfg[cfg_var], QPyNullVariant):
			return True
		return self.plugin not in cfg[cfg_var]

	def toggle(self):

		"""
		desc:
			Toggles the enabled status of the plugin.
		"""

		cfg_var = u'disabled_%s' % self.info[u'type']
		disabled_plugins = cfg[cfg_var].split(u';')
		if self.ui.checkbox_enable.isChecked():
			if self.plugin in disabled_plugins:
				disabled_plugins.remove(self.plugin)
		else:
			if self.plugin not in disabled_plugins:
				disabled_plugins.append(self.plugin)
		cfg[cfg_var] = u';'.join(disabled_plugins)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""


from libopensesame.py3compat import *
from openexp._canvas._element.element import Element
from openexp.canvas_elements import ElementFactory


class Group(Element):

	def __init__(self, canvas, elements=[], **properties):

		if not all(isinstance(e, (Element, ElementFactory)) for e in elements):
			raise TypeError('A group can only contain canvas elements')
		self._elements = [e if isinstance(e, Element) else e.construct(canvas) \
			for e in elements]
		Element.__init__(self, canvas, **properties)

	def prepare(self):

		for element in self._elements:
			element.prepare()

	def show(self, **kwargs):

		for element in self._elements:
			element.show(**kwargs)

	def __contains__(self, xy):

		return any(xy in element for element in self._elements)

	def __iter__(self):

		for e in self._elements:
			yield e

	def __len__(self):

		return len(self._elements)

	def __add__(self, element):

		return Group(self._canvas, self._elements + [element])

	@staticmethod
	def _setter(key, self, val):

		if key == u'visible':
			return Element._setter(key, self, val)
		raise TypeError(u'Properties of element groups cannot be changed')



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception


class response_info(object):

	"""
	desc:
		A single response.
	"""

	def __init__(self, response_store, response=None, correct=None,
		response_time=None, item=None, feedback=True):

		self._response_store = response_store

		# Sanity checks
		if not isinstance(response_time, (int, float)) and \
			response_time is not None:
			raise osexception(u'response should be a numeric value or None')
		if response_time is None:
			self.response_time = None
		else:
			self.response_time = float(response_time)
		self.response = response
		self.item = item
		self.feedback = feedback
		if correct not in (0, 1, True, False, None):
			raise osexception(
				u'correct should be 0, 1, True, False, or None')
		if correct is None:
			self.correct = None
		else:
			self.correct = int(correct)

	def match(self, **kwdict):

		for key, val in kwdict.items():
			if getattr(self, key) != val:
				return False
		return True

	def matchnot(self, **kwdict):

		for key, val in kwdict.items():
			if getattr(self, key) == val:
				return False
		return True

	def __str__(self):

		return u'response(response=%s,response_time=%s,correct=%s,item=%s,feedback=%s)' \
			% (self.response, self.response_time, self.correct, self.item, \
			self.feedback)


class response_store(object):

	"""
	desc: |
		The `responses` object contains the history of the responses that were
		collected during the experiment.

		A `responses` object is created automatically when the experiment starts.

		In addition to the functions listed below, the following semantics are
		supported:

		__Example__:

		~~~ .python
		# Loop through all responses, where last-given responses come first
		# Each response has correct, response, response_time, item, and feedback
		# attributes.
		for response in responses:
			print(response.correct)
		# Print the two last-given respones
		print('last_two responses:')
		print(responses[:2])
		~~~

		[TOC]
	"""

	def __init__(self, experiment):

		"""
		visible: False

		desc:
			Constructor.

		arguments:
			experiment:
				desc:	The experiment object.
				type:	experiment.
		"""

		self._experiment = experiment
		self._responses = []
		self._feedback_from = 0

	@property
	def acc(self):

		"""
		name: acc

		desc:
			The percentage of correct responses for all responses that are included
			in feedback. If there are no responses to give feedback on,
			'undefined' is returned.

		example: |
			print('The accuracy was %s%%' % responses.acc)
		"""

		l = self._select(feedback=True)._selectnot(correct=None).correct
		if not l:
			return u'undefined'
		return 100.*sum(l)/len(l)

	@property
	def avg_rt(self):

		"""
		name: avg_rt

		desc:
			The average response time for all responses that are included
			in feedback. If there are no responses to give feedback on,
			'undefined' is returned.

		example: |
			print('The average RT was %s ms' % responses.avg_rt)
		"""

		l = self._select(feedback=True)._selectnot(response_time=None) \
			.response_time
		if not l:
			return u'undefined'
		return 1.*sum(l)/len(l)

	@property
	def response(self):

		"""
		name: response

		desc:
			A list of all response values. (I.e. not response objects, but
			actual response keys, buttons, etc.)

		example: |
			for response in responses.response:
				print(response)
		"""

		return [r.response for r in self._responses]

	@property
	def correct(self):

		"""
		name: correct

		desc:
			A list of all correct (0, 1, or None) values.

		example: |
			for correct in responses.correct:
				print(correct)
		"""

		return [r.correct for r in self._responses]

	@property
	def response_time(self):

		"""
		name: response_time

		desc:
			A list of all response times (float or None).

		example: |
			for rt in responses.response_time:
				print(rt)
		"""

		return [r.response_time for r in self._responses]

	@property
	def item(self):

		"""
		name: item

		desc:
			A list of all item names (str or None) associated with each
			response.

		example: |
			for item in responses.item:
				print(item)
		"""

		return [r.item for r in self._responses]

	@property
	def feedback(self):

		"""
		name: feedback

		desc:
			A list of the feedback status (True or False) associated with each
			response.

		example: |
			for feedback in responses.feedback:
				print(feedback)
		"""

		return [r.feedback for r in self._responses]

	@property
	def var(self):

		return self._experiment.var

	def add(self, response=None, correct=None, response_time=None, item=None,
		feedback=True):

		"""
		desc:
			Adds a response.

		keywords:
			response:
				desc:	The response value, for example, 'space' for the
						spacebar, 0 for joystick button 0, etc.
			correct:
				desc:	The correctness of the response.
				type:	[bool, int, None]
			response_time:
				desc:	The response_time.
				type:	[float, int, None]
			item:
				desc:	The item that collected the response.
				type:	[str, None]
			feedback:
				desc:	Indicates whether the response should be included in
						feedback on accuracy and average response time.
				type:	bool

		example: |
			responses.add(response_time=500, correct=1, response='left')
		"""

		r = response_info(self, response=response, correct=correct,
			response_time=response_time, item=item, feedback=feedback)
		if correct is None:
			correct = u'undefined'
		else:
			correct = r.correct
		self._responses.insert(0, r)
		self.var.response = self._experiment.syntax.sanitize(r.response)
		self.var.response_time = r.response_time
		self.var.correct = correct
		if item is not None:
			self.var.set(u'response_%s' % item,
				self._experiment.syntax.sanitize(r.response))
			self.var.set(u'response_time_%s' % item, r.response_time)
			self.var.set(u'correct_%s' % item, correct)
		self.var.acc = self.var.accuracy = self.acc
		self.var.avg_rt = self.avg_rt
		# Old variables, mostly for backwards compatibility
		rs = self._select(feedback=True)
		self.var.accuracy = self.var.acc
		self.var.average_response_time = self.var.avg_rt
		self.var.total_response_time = sum(
			rs._selectnot(response_time=None).response_time)
		self.var.total_responses = len(rs)
		self.var.total_correct = len(rs._select(correct=1))

	def clear(self):

		"""
		desc:
			Clears all responses.

		example: |
			responses.clear()
		"""

		self._responses = []

	def reset_feedback(self):

		"""
		desc:
			Sets the feedback status of all responses to False, so that only
			new responses will be included in feedback.

		example: |
			responses.reset_feedback()
		"""

		for r in self._responses:
			r.feedback = False

	def _select(self, **kwdict):

		rs = response_store(self._experiment)
		for r in self._responses:
			if r.match(**kwdict):
				rs._responses.append(r)
		return rs

	def _selectnot(self, **kwdict):

		rs = response_store(self._experiment)
		for r in self._responses:
			if r.matchnot(**kwdict):
				rs._responses.append(r)
		return rs

	def __len__(self):

		return len(self._responses)

	def __getitem__(self, key):

		if isinstance(key, slice):
			rs = response_store(self._experiment)
			rs._responses = self._responses[key]
			return rs
		if isinstance(key, int):
			return self._responses[key]
		raise TypeError(u'A key for responses should be either slice or int')

	def __str__(self):

		s = u'%d responses (last response is shown first):\n' % len(self)
		for i, r in enumerate(self):
			s += '%d: %s\n' % (i, r)
		return s

	def __iter__(self):

		for r in self._responses:
			yield r



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import base_extension


class restore_ui_elements(base_extension):

	"""
	desc:
		Restores key UI elements when they have been accidentally hidden.
	"""

	@base_extension.as_thread(1000)
	def event_startup(self):
		
		if (
			self.main_window.toolbar_items.isVisible() and
			(self.menubar.isVisible() or self.toolbar.isVisible()) and
			self.main_window.dock_overview.isVisible()
		):
			return
		oslogger.debug('some key ui elements are hidden')
		from libqtopensesame._input.confirmation import confirmation
		if not confirmation(
			self.main_window,
			_('Some important elements of the user interface have been hidden. Do you want to show them again?'),
			title=_('Restore user interface'),
			default='yes'
		).show():
			return
		oslogger.debug('showing key ui elements')
		self.main_window.toolbar_items.setVisible(True)
		self.menubar.setVisible(True)
		self.toolbar.setVisible(True)
		self.main_window.dock_overview.setVisible(True)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.misc.sketchpad_canvas import sketchpad_canvas
from libqtopensesame.widgets.sketchpad_widget import sketchpad_widget
from libqtopensesame.items.qtplugin import qtplugin
from libqtopensesame import sketchpad_elements
from libqtopensesame.validators import duration_validator

class feedpad(object):

	"""
	desc:
		Controls for the sketchpad item.
	"""

	help_url = u'manual/stimuli/visual'
	lazy_init = True

	def init_edit_widget(self):

		"""
		desc:
			Initializes the widget.
		"""

		qtplugin.init_edit_widget(self, False)
		self.canvas = sketchpad_canvas(self)
		self.sketchpad_widget = sketchpad_widget(self)
		self.add_widget(self.sketchpad_widget)
		self.auto_add_widget(self.sketchpad_widget.ui.edit_duration,
			u'duration')
		self.sketchpad_widget.ui.edit_duration.setValidator(
			duration_validator(self))
		self.first_refresh = True
		self._lock = False

	def apply_script_changes(self):

		"""See qtitem."""

		# Without this hack, the script changes will automatically trigger a
		# draw(), which triggers a call to apply_edit_changes, which updates the
		# script. To avoid this feedback loop, temporarily disable apply edit
		# changes.
		f = self.apply_edit_changes
		self.apply_edit_changes = lambda: None
		qtplugin.apply_script_changes(self)
		self.apply_edit_changes = f

	def element_module(self):

		"""
		returns:
			desc:	The module that contains the sketchpad elements.
			type:	module
		"""

		return sketchpad_elements

	def set_focus(self):

		"""
		desc:
			Allows the item to focus the most important widget.
		"""

		self.sketchpad_widget.ui.edit_duration.setFocus()

	def edit_widget(self):

		"""
		desc:
			Updates the widget.
		"""

		qtplugin.edit_widget(self)
		self.sketchpad_widget.initialize()
		self.sketchpad_widget.draw()
		if self.first_refresh:
			self.sketchpad_widget.center()
			self.sketchpad_widget.zoom_fit()
			self.first_refresh = False

	def add_element(self, element):

		"""
		desc:
			Adds an element to the sketchpad.

		arguments:
			element:
				desc:	The element to add.
				type:	base_element
		"""

		if element is None:
			return
		self.elements.append(element)
		self.draw()
		for element in self.elements:
			element.select(False)
		element.select()

	def selected_elements(self):

		"""
		desc:
			Gets selected elements.

		returns:
			desc:	A list of selected elements.
			type:	list
		"""

		l = []
		for element in self.elements:
			if element.selected:
				l.append(element)
		return l

	def remove_elements(self, elements):

		"""
		desc:
			Removes elements.

		arguments:
			elements:
				desc:	A list of elements to be removed.
				type:	list
		"""

		for element in elements:
			self.elements.remove(element)

	def move_elements(self, elements, dx=0, dy=0):

		"""
		desc:
			Moves elements.

		arguments:
			elements:
				desc:	A list of elements to be moved.
				type:	list

		keywords:
			dx:
				desc:	A horizontal displacement.
				type:	[int, float]
			dy:
				desc:	A vertical displacement
				type:	[int, float]
		"""

		for element in elements:
			element.move(dx, dy)

	def topleft_elements_pos(self, elements):

		"""
		desc:
			Gets the top-left position of a set of elements.

		arguments:
			elements:
				desc:	A list of elements.
				type:	list

		returns:
			desc:		An (x,y) tuple with the top-left position.
			type:		tuple.
		"""

		lx = []
		ly = []
		for element in elements:
			x, y = element.get_pos()
			lx.append(x)
			ly.append(y)
		return min(lx), min(ly)

	def set_elements_pos(self, elements, x=0, y=0):

		"""
		desc:
			Sets position of elements.

		arguments:
			elements:
				desc:	A list of elements to be positioned.
				type:	list

		keywords:
			x:
				desc:	A horizontal position.
				type:	[int, float]
			y:
				desc:	A vertical position
				type:	[int, float]
		"""

		mx, my = self.topleft_elements_pos(elements)
		for element in elements:
			dx, dy = element.get_pos()
			dx -= mx
			dy -= my
			element.set_pos(x+dx, y+dy)

	def min_z_index(self):

		"""
		desc:
			Gets the z-index of the top item (i.e. in front of all other items).

		returns:
			desc:	A z-index.
			type:	int
		"""

		return min([
			element.z_index
			for element in self.elements
			if isinstance(element.z_index, (int, float))
		])

	def max_z_index(self):

		"""
		desc:
			Gets the z-index of the bottom item (i.e. below all other items).

		returns:
			desc:	A z-index.
			type:	int
		"""

		return max([
			element.z_index
			for element in self.elements
			if isinstance(element.z_index, (int, float))
		])

	@property
	def current_color(self):
		return self.sketchpad_widget.current_color

	@property
	def current_penwidth(self):
		return self.sketchpad_widget.current_penwidth

	@property
	def current_arrow_head_width(self):
		return self.sketchpad_widget.current_arrow_head_width

	@property
	def current_arrow_body_width(self):
		return self.sketchpad_widget.current_arrow_body_width

	@property
	def current_arrow_body_length(self):
		return self.sketchpad_widget.current_arrow_body_length

	@property
	def current_scale(self):
		return self.sketchpad_widget.current_scale

	@property
	def current_rotation(self):
		return self.sketchpad_widget.current_rotation

	@property
	def current_show_if(self):
		return self.sketchpad_widget.current_show_if

	@property
	def current_fill(self):
		return self.sketchpad_widget.current_fill

	@property
	def current_center(self):
		return self.sketchpad_widget.current_center

	@property
	def current_font_family(self):
		return self.sketchpad_widget.current_font_family

	@property
	def current_font_size(self):
		return self.sketchpad_widget.current_font_size

	@property
	def current_font_bold(self):
		return self.sketchpad_widget.current_font_bold

	@property
	def current_font_italic(self):
		return self.sketchpad_widget.current_font_italic

	@property
	def current_html(self):
		return self.sketchpad_widget.current_html

	@property
	def draw(self):
		return self.sketchpad_widget.draw

	@property
	def set_cursor_pos(self):
		return self.sketchpad_widget.set_cursor_pos

	@property
	def graphics_view(self):
		return self.sketchpad_widget.ui.graphics_view

	@property
	def zoom_diff(self):
		return self.sketchpad_widget.zoom_diff

	@property
	def show_element_settings(self):
		return self.sketchpad_widget.show_element_settings

	@property
	def select_pointer_tool(self):
		return self.sketchpad_widget.select_pointer_tool



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._coordinates.coordinates import Coordinates


class Psycho(Coordinates):

	"""
	desc:
		For function specifications and docstrings, see
		`openexp._coordinates.coordinates`.
	"""

	def to_xy(self, x, y=None):

		if isinstance(x, tuple):
			x, y = x
		x, y = self.none_to_center(x, y)
		# For PsychoPy, 0,0 is the display center and positive y
		# coordinates are down.
		if self.uniform_coordinates:
			return x, -y
		return x - self._xcenter, self._ycenter - y

	def from_xy(self, x, y=None):

		if y is None:
			x, y = x
		if self.uniform_coordinates:
			return x, -y
		return x + self._xcenter, self._ycenter - y


# Non PEP-8 alias for backwards compatibility
psycho = Psycho



#-*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.misc.config import cfg
from libqtopensesame.sketchpad_elements._base_element import base_element

class base_line_arrow(base_element):

	"""
	desc:
		A base class for line and arrow elements.

		See base_element for docstrings and function descriptions.
	"""

	pos_start = None
	preview = None

	def set_pos(self, x=0, y=0):

		dx = self.properties[u'x2'] - self.properties[u'x1']
		dy = self.properties[u'y2'] - self.properties[u'y1']
		self.properties[u'x1'] = x
		self.properties[u'y1'] = y
		self.properties[u'x2'] = x+dx
		self.properties[u'y2'] = y+dy

	@classmethod
	def mouse_press(cls, sketchpad, pos):

		if cls.pos_start is not None:
			return
		cls.pos_start = pos
		cls.preview = sketchpad.canvas.line(pos[0], pos[1], pos[0], pos[1],
			color=cfg.sketchpad_preview_color,
			penwidth=cfg.sketchpad_preview_penwidth)

	@classmethod
	def mouse_move(cls, sketchpad, pos):

		if cls.pos_start is None or cls.preview is None:
			return
		cls.preview.setLine(cls.pos_start[0], cls.pos_start[1], pos[0], pos[1])

	@classmethod
	def reset(cls):

		cls.pos_start = None

	@staticmethod
	def requires_color():
		return True

	@staticmethod
	def requires_penwidth():
		return True



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.extensions import base_extension
from tab_to_dockwidget_docktab import DockTab
from qtpy import QtCore


class tab_to_dockwidget(base_extension):

	"""
	desc:
		This extensions turns tabs into standalone dockwidgets, so that you can
		see multiple things side by side.
	"""

	def event_startup(self):

		"""
		desc:
			During initialization we monkey-patch the tabwidget so that it also
			manages dockwidgets.
		"""

		self._docktabs = {}
		self.overview_area.structure_change.connect(self.update_all)
		self.tabwidget.add = self._add_tab(self.tabwidget.add)
		self.tabwidget.remove = self._remove_tab(self.tabwidget.remove)
		self.tabwidget.close_all = self._close_all_tabs(self.tabwidget.close_all)

	def event_run_experiment(self, fullscreen):

		"""
		desc:
			Disable all docked item controls when running the experiment.
		"""

		for docktab in self._docktabs.values():
			docktab.setEnabled(False)

	def event_end_experiment(self, ret_val):

		"""
		desc:
			Re-enable all docked item controls when the experiment is done.
		"""

		for docktab in self._docktabs.values():
			docktab.setEnabled(True)

	def update_all(self):

		"""
		desc:
			Update all item controls.
		"""

		for docktab in self._docktabs.values():
			docktab.update()

	def event_rename_item(self, from_name, to_name):

		"""
		desc:
			Change the dockwidget titles on item rename, and allow items to
			update in response to renames.
		"""

		for docktab in self._docktabs.values():
			docktab.rename(from_name, to_name)

	def event_new_item(self, name, _type):

		"""
		desc:
			Allow items to update in response to new items.
		"""

		self.update_all()

	def event_delete_item(self, name):

		"""
		desc:
			Allow items to update in response to item deletions.
		"""

		for docktab in self._docktabs.values():
			docktab.delete(name)

	def activate(self):

		"""
		desc:
			Turns the currently visible tab (if any) into a dockwidget.
		"""

		widget = self.tabwidget.currentWidget()
		if widget is None:
			return
		name = self.tabwidget.tabText(self.tabwidget.currentIndex())
		self.tabwidget.remove(widget)
		docktab = DockTab(self, widget, name)
		self._docktabs[widget] = docktab
		self.main_window.addDockWidget(QtCore.Qt.RightDockWidgetArea, docktab)

	def remove_widget(self, widget):

		"""
		desc:
			Removes a docked widget.
		"""

		if widget in self._docktabs:
			del self._docktabs[widget]

	def _close_all_tabs(self, fnc):

		"""
		desc:
			Decorates tabwidget.close_all()
		"""

		def inner(*args, **kwargs):

			while self._docktabs:
				widget, docktab = self._docktabs.popitem()
				docktab.close()
			return fnc(*args, **kwargs)

		return inner

	def _add_tab(self, fnc):

		"""
		desc:
			Decorates tabwidget.add()
		"""

		def inner(widget, *args, **kwargs):

			if widget in self._docktabs:
				if hasattr(widget, u'set_focus'):
					widget.set_focus()
				return
			return fnc(widget, *args, **kwargs)

		return inner

	def _remove_tab(self, fnc):

		"""
		desc:
			Decorates tabwidget.remove()
		"""

		def inner(widget, *args, **kwargs):

			if widget in self._docktabs:
				docktab = self._docktabs.pop(widget)
				docktab.close()
				return
			return fnc(widget, *args, **kwargs)

		return inner



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import warnings
import random
import itertools
import math
from libopensesame.exceptions import osexception
from openexp.backend import Backend, configurable
from openexp.color import Color
from PIL import Image as PILImage
from collections import OrderedDict
from openexp.canvas_elements import (
	Line,
	Rect,
	Polygon,
	Ellipse,
	Image,
	Gabor,
	NoisePatch,
	Circle,
	FixDot,
	ElementFactory,
	RichText,
	Arrow
)
from openexp._canvas._element.element import Element
from openexp._canvas._element.group import Group


class Canvas(Backend):

	"""
	desc: |
		The `Canvas` class is used to present visual stimuli. You generally
		create a `Canvas` object with the `Canvas()` factory function, as
		described in the section [Creating a Canvas](#creating-a-canvas).

		__Example__:

		~~~ .python
		# Create and show a canvas with a central fixation dot
		my_canvas = Canvas()
		my_canvas.fixdot()
		my_canvas.show()
		~~~

		__Example__:

		As of OpenSesame 3.2, you can also add `Canvas` elements as objects.
		See also the section on [Naming, accessing, and modifying elements](#naming-accessing-and-modifying-elements).

		~~~ .python
		# Create a canvas with a fixation dot and a rectangle
		my_canvas = Canvas()
		my_canvas['my_fixdot'] = FixDot()
		my_canvas.show()
		~~~

		[TOC]

		## Things to know

		### Creating a Canvas

		You generally create a `Canvas` with the `Canvas()` factory function:

		~~~ .python
		my_canvas = Canvas()
		~~~

		Optionally, you can pass [Style keywords](#style-keywords) to `Canvas()`
		to set the default style:

		~~~ .python
		my_canvas = Canvas(color='green')
		my_canvas.fixdot() # Will be green
		~~~

		### Style keywords

		All functions that accept `**style_args` take the following keyword
		arguments:

		- `color` specifies the foreground color. For valid color
			specifications, see [colors].
		- `background_color` specifies the background color. For valid color
			specifications, see [colors].
		- `fill` indicates whether rectangles, circles, polygons, and ellipses
			are filled (`True`), or drawn as an outline (`False`).
		- `penwidth` indicates a penwidth in pixels and should be `int` or
			`float`.
		- `html` indicates whether HTML-tags are interpreted, and should be
			`True` or `False`. For supported tags, see [/usage/text/]().
		- `font_family` is the name of a font family, such as 'sans'.
		- `font_size` is a font size in pixels.
		- `font_italic` indicates whether text should italics, and should be
			`True` or `False`.
		- `font_bold` indicates whether text should bold, and should be
			`True` or `False`.
		- `font_underline` indicates whether text should underlined, and should
			be `True` or `False`.

		~~~ .python
		# Draw a green fixation dot
		my_canvas = Canvas()
		my_canvas.fixdot(color='green')
		my_canvas.show()
		~~~

		Style keywords only affect the current drawing operation (except when
		passed to `Canvas()` while creating the `Canvas`). To change the style
		for all subsequent drawing operations, set style properties, such as
		[canvas.color], directly:

		~~~ .python
		# Draw a red cross with a 2px penwidth
		my_canvas = Canvas()
		my_canvas.color = u'red'
		my_canvas.penwidth = 2
		my_canvas.line(-10, -10, 10, 10)
		my_canvas.line(-10, 10, 10, -10)
		my_canvas.show()
		~~~

		Or pass the style properties to `Canvas()`:

		~~~ .python
		# Draw a red cross with a 2px penwidth
		my_canvas = Canvas(color=u'red', penwidth=2)
		my_canvas.line(-10, -10, 10, 10)
		my_canvas.line(-10, 10, 10, -10)
		my_canvas.show()
		~~~


		### Colors

		You can specify colors in the following ways. This includes CSS3-type
		color specifications, but also supports some extra specifications,
		such as CIE l\* a\* b\* color space.

		__Version note:__ The `hsv`, `hsl`, and `lab` color spaces are new in
		v3.3.0.

		- __Color names:__ 'red', 'black', etc. A full list of valid color names
			can be found
			[here](http://www.w3.org/TR/SVG11/types.html#ColorKeywords).
		- __Seven-character hexadecimal strings:__ `#FF0000`, `#000000`, etc.
			Here, values range from `00` to `FF`, so that `#FF0000` is bright red.
		- __Four-character hexadecimal strings:__ `#F00`, `#000`, etc. Here,
			values range from '0' to 'F' so that `#F00` is bright red.
		- __RGB strings:__ `rgb(255,0,0)`, `rgb(0,0,0)`, etc. Here, values
			range from 0 to 255 so that `rgb(255,0,0)` is bright red.
		- __RGB percentage strings:__ `rgb(100%,0%,0%)`, `rgb(0%,0%,0%)`, etc.
			Here, values range from 0% to 100% so that `rgb(100%,0%,0%)` is
			bright red.
		- __RGB tuples:__ `(255, 0, 0)`, `(0, 0 ,0)`, etc. Here, values range
			from `0` to `255` so that `(255,0,0)' is bright red.
		- __HSV strings:__ `hsv(120, 100%, 100%)`. In the [HSV](https://en.wikipedia.org/wiki/HSL_and_HSV)
			color space, the hue parameter is an angle from 0 to 359, and the
			saturation and value parameters are percentages from 0% to 100%.
		- __HSL strings:__ `hsl(120, 100%, 50%)`. In the [HSL](https://en.wikipedia.org/wiki/HSL_and_HSV)
			color space, the hue parameter is an angle from 0 to 359, and the
			saturation and lightness parameters are percentages from 0% to 100%.
		- __LAB strings:__ `lab(53, -20, 0)`. In the [CIELAB](https://en.wikipedia.org/wiki/CIELAB_color_space)
			color space, the parameters reflect lightness (`l*`),
			green-red axis (`a*`, negative is green), and blue-yellow axis
			(`b*`, negative is blue). This uses the D65 white point and the
			sRGB transfer function, as implemented [here](https://www.psychopy.org/_modules/psychopy/tools/colorspacetools.html).
		- __Luminance values:__  `255`, `0`, etc. Here, values range from `0` to
			`255` so that `255` is white.

		~~~ .python
		# Various ways to specify green
		my_canvas.fixdot(color='green')  # Dark green
		my_canvas.fixdot(color='#00ff00')
		my_canvas.fixdot(color='#0f0')
		my_canvas.fixdot(color='rgb(0, 255, 0)')
		my_canvas.fixdot(color='rgb(0%, 100%, 0%)')
		my_canvas.fixdot(color='hsl(100, 100%, 50%)')
		my_canvas.fixdot(color='hsv(0, 100%, 100%)')
		my_canvas.fixdot(color='lab(53, -20, 0)')  # Dark green
		my_canvas.fixdot(color=(0, 255, 0))
		# Specify a luminance value (white)
		my_canvas.fixdot(color=255)
		~~~


		### Naming, accessing, and modifying elements

		As of OpenSesame 3.2, the `Canvas` supports an object-based interface
		that allows you to name elements, and to access and modify elements
		individually, without having to redraw the entire `Canvas`.

		For example, the following will first add a red `Line` element to a
		`Canvas` and show it, and then change the color of the line to green and
		show it again. The name of the element (`my_line`) is used to refer to
		the element later on to change it.

		~~~ .python
		my_canvas = Canvas()
		my_canvas['my_line'] = Line(-100, -100, 100, 100, color='red')
		my_canvas.show()
		clock.sleep(1000)
		my_canvas['my_line'].color = 'green'
		my_canvas.show()
		~~~

		You can also add an element without explicitly providing a name for it.
		In that case, a name is generated automatically (e.g. `stim0`).

		~~~ .python
		my_canvas = Canvas()
		my_canvas += FixDot()
		my_canvas.show()
		~~~

		If you add a list of elements, they will be automatically grouped
		together, and you can refer to the entire group by name.

		~~~ .python
		my_canvas = Canvas()
		my_canvas['my_cross'] = [
			Line(-100, 0, 100, 0),
			Line(0, -100, 0, 100)
		]
		my_canvas.show()
		~~~

		To check whether a particular `x,y` coordinate falls within the bounding
		rectangle of an element, you can use `in`:

		~~~ .python
		my_mouse = Mouse(visible=True)
		my_canvas = Canvas()
		my_canvas['rect'] = Rect(-100, -100, 200, 200)
		my_canvas.show()
		button, (x, y), time = my_mouse.get_click()
		if (x, y) in my_canvas['rect']:
			print('Clicked in rectangle')
		else:
			print('Clicked outside of rectangle')
		~~~

		You can also get a list of the names of all elements that contain an
		`x,y` coordinate, using the `Canvas.elements_at()` function, documented
		below.


		%--
		constant:
			arg_max_width: |
				The maximum width of the text in pixels, before wrapping to a
				new line, or `None` to wrap at screen edge.
			arg_bgmode: |
				Specifies whether the background is the average of col1 col2
				('avg', corresponding to a typical Gabor patch), or equal to
				col2 ('col2'), useful for blending into the	background. Note:
				this parameter is ignored by the psycho backend, which uses
				increasing transparency for the background.
			arg_style: |
				Optional [style keywords] that specify the style of the current
				drawing operation. This does not affect subsequent drawing
				operations.
			arg_center: |
				A bool indicating whether the coordinates reflect the center
				(`True`) or top-left (`False`) of the text.
		--%
	"""

	# The maximum duration that Canvas.show() should take, which is set to one
	# frame duration on a 60 Hz monitor. If Canvas.show() takes longer, a
	# warning given.
	MAX_SHOW_DT = 16

	def __init__(self, experiment, auto_prepare=True, **style_args):

		"""
		visible: False

		desc: |
			Constructor to create a new `Canvas` object. You do not generally
			call this constructor directly, but use the `Canvas()` factory
			function, which is described here: [/python/common/]().

		arguments:
			experiment:
				desc:	The experiment object.
				type:	experiment

		keywords:
			auto_prepare:
				desc:	Indicates whether the canvas should be automatically
						prepared after each drawing operation, so that
						[canvas.show] will be maximally efficient. If
						auto_prepare is turned off, drawing operations may
						be faster, but [canvas.show] will take longer,
						unless [canvas.prepare] is explicitly called in
						advance. This option exists mostly for historical
						purposes, because there are currently no backends for
						which it is necessary to disable auto prepare.
				type:	bool

		keyword-dict:
			style_args:
				Optional [style keywords], which will be used as the default
				for all drawing operations on this `Canvas`.

		example: |
			my_canvas = Canvas()
			my_canvas.fixdot()
			my_canvas.show()
		"""

		self.experiment = experiment
		self._width = self.experiment.var.width
		self._height = self.experiment.var.height
		self.auto_prepare = auto_prepare
		Backend.__init__(self, configurables={
			u'color' : None,
			u'background_color' : None,
			u'fill' : self.assert_bool,
			u'penwidth' : self.assert_numeric,
			u'html' : self.assert_bool,
			u'font_family' : self.assert_string,
			u'font_size' : self.assert_numeric,
			u'font_italic' : self.assert_bool,
			u'font_bold' : self.assert_bool,
			u'font_underline' : self.assert_bool,
		}, **style_args)
		self._elements = OrderedDict()
		self._stimnr = 0

	def __enter__(self):

		"""
		visible: False

		desc:
			The context manager provides an elegant way to disable auto
			preparation. However, it is hidden for now because none of the
			backends require it.
		"""

		self._original_auto_prepare = self.auto_prepare
		self.auto_prepare = False
		return self

	def __exit__(self, exc_type, exc_value, traceback):

		"""
		visible: False

		desc:
			The context manager provides an elegant way to disable auto
			preparation. However, it is hidden for now because none of the
			backends require it.
		"""

		self.auto_prepare = self._original_auto_prepare
		self.prepare()

	def __setitem__(self, key, value):

		"""
		visible: False

		desc:
			Sets a named element. If a list of elements is provided, a Group
			is automatically created.
		"""

		if key in self._elements:
			raise ValueError('An element with the name "%s" already exists' %
				key)
		if isinstance(value, list):
			value = Group(self, [e.construct(self)
				if isinstance(e, ElementFactory) else e for e in value])
		if isinstance(value, ElementFactory):
			value = value.construct(self)
		if not isinstance(value, Element):
			raise TypeError('%s is not a canvas element but %s' %
				(key, type(value)))
		self._elements[key] = value

	def __delitem__(self, key):

		"""
		visible: False

		desc:
			Deletes an element by name.
		"""

		del self._elements[key]

	def __getitem__(self, key):

		"""
		visible: False

		desc:
			Retrieves an element by name.
		"""

		return self._elements[key]

	def __len__(self):

		"""
		visible: False

		desc:
			Gets the number of elements, where a Group counts as 1.
		"""

		return len(self._elements)

	def __contains__(self, key):

		"""
		visible: False

		desc:
			Checks whether a named element is in the Canvas.
		"""

		return key in self._elements

	def __iter__(self):

		"""
		visible: False

		desc:
			Iterates over the elements.
		"""

		for name, stim in self._elements.items():
			yield name, stim

	def __iadd__(self, element):

		"""
		visible: False

		desc:
			Adds an unnamed element.
		"""

		while 'stim%d' % self._stimnr in self._elements:
			self._stimnr += 1
		self['stim%d' % self._stimnr] = element
		return self

	def _get_name_element(self, element):

		"""
		visible: False

		desc:
			Gets a name, element tuple for an element or element name (i.e.
			get both if you have only one).

		arguments:
			element:
				desc:	A SKETCHPAD element, or its name.
				type:	[Element, str]

		returns:
			A (name, element) tuple.
		"""

		if isinstance(element, basestring):
			return element, self._elements[element]
		for name, _element in self._elements.items():
			if element is _element:
				return name, element
		raise ValueError('"%s" not found in canvas"' % element)

	def elements_at(self, x, y):

		"""
		desc: |
			*New in v3.2.0*

			Gets the names of elements that contain a particular `x, y`
			coordinate.

		arguments:
			x:
				desc:	An X coordinate.
				type:	[int, float]
			y:
				desc:	A Y coordinate.
				type:	[int, float]

		returns:
			desc:	A `list` of element names that contain the coordinate
					`x, y`.
			type:	list

		example: |
		  # Create and show a canvas with two partly overlapping rectangles
		  my_canvas = Canvas()
		  my_canvas['right_rect'] = Rect(x=-200, y=-100, w=200, h=200, color='red')
		  my_canvas['left_rect'] = Rect(x=-100, y=-100, w=200, h=200, color='green')
		  my_canvas.show()
		  # Collect a mouse click and print the names of the elements that
		  # contain the clicked point
		  my_mouse = Mouse(visible=True)
		  button, pos, time = my_mouse.get_click()
		  if pos is not None:
		      x, y = pos
		      print('Clicked on elements: %s' % my_canvas.elements_at(x, y))
		"""

		elements = []
		for name, element in self:
			try:
				if (x, y) in element:
					elements.append(name)
			except NotImplementedError:
				pass
		return elements

	def lower_to_bottom(self, element):

		"""
		desc:
			Lowers an element to the bottom, so that it is drawn first; that is,
			it becomes the background.

		arguments:
			element:
				desc:	A SKETCHPAD element, or its name.
				type:	[Element, str]
		"""

		first_name, first_element = self._get_name_element(element)
		self._elements.pop(first_name)
		_elements = OrderedDict()
		_elements[first_name] = first_element
		for name, element in self._elements.items():
			_elements[name] = element
		self._elements = _elements

	def raise_to_top(self, element):

		"""
		desc:
			Raises an element to the top, so that it is drawn last; that is,
			it becomes the foreground.

		arguments:
			element:
				desc:	A SKETCHPAD element, or its name.
				type:	[Element, str]
		"""

		last_name, last_element = self._get_name_element(element)
		self._elements.pop(last_name)
		_elements = OrderedDict()
		for name, element in self._elements.items():
			_elements[name] = element
		_elements[last_name] = last_element
		self._elements = _elements

	def add_element(self, element, name=None):

		"""
		visible: False

		desc:	An alternative to the dict and += API. For internal use.
		"""

		if name is None:
			self.__iadd__(element)
			return
		self._elements[name] = element

	def rename_element(self, old_name, new_name):

		"""
		visible: False

		desc: 	Renames an element.
		"""

		element = self._elements.pop(old_name)
		i = 1
		name = new_name
		while name in self._elements:
			name = '%s_%d' % (new_name, i)
			i += 1
		self._elements[name] = element

	def set_config(self, **cfg):

		# Remap deprecated names
		if u'bgcolor' in cfg:
			warnings.warn(u'bgcolor is a deprecated style argument. '
				'Use background_color instead.',
				DeprecationWarning)
			cfg[u'background_color'] = cfg[u'bgcolor']
			del cfg['bgcolor']
		if u'fgcolor' in cfg:
			warnings.warn(u'fgcolor is a deprecated style argument. '
				'Use color instead.', DeprecationWarning)
			cfg[u'color'] = cfg[u'fgcolor']
			del cfg['fgcolor']
		if u'font_style' in cfg:
			warnings.warn(u'font_style is a deprecated style argument. '
				'Use font_family instead.',	DeprecationWarning)
			cfg[u'font_family'] = cfg[u'font_style']
			del cfg['font_style']
		# Convert color to backend specific colors
		if u'color' in cfg and not hasattr(cfg[u'color'], u'backend_color'):
			cfg[u'color'] = Color(self.experiment, cfg[u'color'])
		if u'background_color' in cfg \
			and not hasattr(cfg[u'background_color'], u'backend_color'):
			cfg[u'background_color'] = Color(self.experiment,
				cfg[u'background_color'])
		Backend.set_config(self, **cfg)

	def default_config(self):

		return {
			u'penwidth' 		: 1,
			u'fill'				: False,
			u'html'				: True,
			u'color'			: self.experiment.var.foreground,
			u'background_color'	: self.experiment.var.background,
			u'font_size'		: self.experiment.var.font_size,
			u'font_family'		: self.experiment.var.font_family,
			u'font_bold'		: self.experiment.var.font_bold==u'yes',
			u'font_italic'		: self.experiment.var.font_italic==u'yes',
			u'font_underline'	: self.experiment.var.font_underline==u'yes',
			}

	@property
	def size(self):

		"""
		name:
			size

		desc:
			The size of the canvas as a `(width, height)` tuple. This is a
			read-only property.
		"""

		return self._width, self._height

	@property
	def width(self):

		"""
		name:
			width

		desc:
			The width of the canvas. This is a read-only property.
		"""

		return self._width

	@property
	def height(self):

		"""
		name:
			height

		desc:
			The height of the canvas. This is a read-only property.
		"""

		return self._height

	@property
	def left(self):

		"""
		name:
			left

		desc:
			The x coordinate of the left edge of the display. This is a
			read-only property.
		"""

		return self._left

	@property
	def right(self):

		"""
		name:
			right

		desc:
			The x coordinate of the right edge of the display. This is a
			read-only property.
		"""

		return self._right

	@property
	def top(self):

		"""
		name:
			top

		desc:
			The y coordinate of the top edge of the display. This is a
			read-only property.
		"""

		return self._top

	@property
	def bottom(self):

		"""
		name:
			bottom

		desc:
			The y coordinate of the bottom edge of the display. This is a
			read-only property.
		"""

		return self._bottom

	def copy(self, canvas):

		"""
		desc: |
			Turns the current `Canvas` into a copy of the passed `Canvas`.

			__Warning:__

			Copying `Canvas` objects can result in unpredictable behavior. In
			many cases, a better solution is to recreate multiple `Canvas`
			objects from scratch, and/ or to use the element interface to
			update `Canvas` elements individually.

		arguments:
			canvas:
				desc:	The `Canvas` to copy.
				type:	canvas

		example: |
			my_canvas = Canvas()
			my_canvas.fixdot(x=100, color='green')
			my_copied_canvas = Canvas()
			my_copied_canvas.copy(my_canvas)
			my_copied_canvas.fixdot(x=200, color="blue")
			my_copied_canvas.show()
		"""

		self._elements = OrderedDict([
			(name, element.copy(self))
			for name, element in canvas._elements.items()
		])
		self.set_config(**canvas.get_config())

	def prepare(self):

		"""
		visible: False

		desc:
			Finishes pending canvas operations (if any), so that a subsequent
			call to [canvas.show] is extra fast. It's only necessary to call
			this function if you have disabled `auto_prepare` in
			[canvas.__init__].
		"""

		for name, element in self._elements.items():
			if element.visible:
				element.prepare()

	def show(self):

		"""
		desc:
			Shows, or 'flips', the canvas on the screen.

		returns:
			desc:
				A timestamp of the time at which the canvas actually appeared on
				the screen, or a best guess if precise temporal information is
				not available. For more information about timing, see
				</misc/timing>. Depending on the back-end the timestamp is an
				`int` or a `float`.
			type:
				[int, float]


		example: |
			my_canvas = Canvas()
			my_canvas.fixdot()
			t = my_canvas.show()
			exp.set('time_fixdot', t)
		"""

		raise NotImplementedError()

	@configurable
	def clear(self, **style_args):

		"""
		desc:
			Clears the canvas with the current background color. Note that it is
			generally faster to use a different canvas for each experimental
			display than to use a single canvas and repeatedly clear and redraw
			it.

		keyword-dict:
			style_args:	"%arg_style"

		example: |
			my_canvas = Canvas()
			my_canvas.fixdot(color='green')
			my_canvas.show()
			sleep(1000)
			my_canvas.clear()
			my_canvas.fixdot(color='red')
			my_canvas.show()
		"""

		raise NotImplementedError()

	def fixdot(self, x=None, y=None, style=u'default', **style_args):

		"""
		desc: |
			Draws a fixation dot. The default style is medium-open.

			- 'large-filled' is a filled circle with a 16px radius.
			- 'medium-filled' is a filled circle with an 8px radius.
			- 'small-filled' is a filled circle with a 4px radius.
			- 'large-open' is a filled circle with a 16px radius and a 2px hole.
			- 'medium-open' is a filled circle with an 8px radius and a 2px hole.
			- 'small-open' is a filled circle with a 4px radius and a 2px hole.
			- 'large-cross' is 16px cross.
			- 'medium-cross' is an 8px cross.
			- 'small-cross' is a 4px cross.

		keywords:
			x:
				desc:	The X coordinate of the dot center, or None to draw a
						horizontally centered dot.
				type:	[int, NoneType]
			y:
				desc:	The Y coordinate of the dot center, or None to draw a
						vertically centered dot.
				type:	[int, NoneType]
			style:
				desc: |
						The fixation-dot style. One of: default, large-filled,
						medium-filled, small-filled, large-open, medium-open,
						small-open, large-cross, medium-cross, or small-cross.
						default equals medium-open.
				type:	[str, unicode]

		keyword-dict:
			style_args:	"%arg_style"

		example: |
			my_canvas = Canvas()
			# Function interface
			my_canvas.fixdot()
			# Element interface
			my_canvas['my_fixdot'] = FixDot()
		"""

		self += FixDot(x=x, y=y, style=style, **style_args)
		return 'stim%d' % self._stimnr

	def circle(self, x, y, r, **style_args):

		"""
		desc:
			Draws a circle.

		arguments:
			x:
				desc:	The center X coordinate of the circle.
				type:	int
			y:
				desc:	The center Y coordinate of the circle.
				type:	int
			r:
				desc:	The radius of the circle.
				type:	int

		keyword-dict:
			style_args:	"%arg_style"

		example: |
			my_canvas = Canvas()
			# Function interface
			my_canvas.circle(100, 100, 50, fill=True, color='red')
			# Element interface
			my_canvas['my_circle'] = Circle(100, 100, 50, fill=True, color='red')
		"""

		self += Circle(x, y, r, **style_args)
		return 'stim%d' % self._stimnr

	def line(self, sx, sy, ex, ey, **style_args):

		"""
		desc:
			Draws a line.

		arguments:
			sx:
				desc:	The left X coordinate.
				type:	int
			sy:
				desc:	The top Y coordinate.
				type:	int
			ex:
				desc:	The right X coordinate.
				type:	int
			ey:
				desc:	The bottom Y coordinate.
				type:	int

		keyword-dict:
			style_args:	"%arg_style"

		Example: |
			my_canvas = Canvas()
			w = self.var.width
			h = self.var.height
			# Function interface
			my_canvas.line(0, 0, w, h)
			# Element interface
			my_canvas['my_line'] = Line(0, 0, w, h)
		"""

		self += Line(sx, sy, ex, ey, **style_args)
		return 'stim%d' % self._stimnr

	def arrow(self, sx, sy, ex, ey, body_length=0.8, body_width=.5,
		head_width=30, **style_args):

		"""
		desc:
			Draws an arrow. An arrow is a polygon consisting of 7 vertices,
			with an arrowhead pointing at (ex, ey).

		arguments:
			sx:
				desc:	The X coordinate of the arrow's base.
				type:	int
			sy:
				desc:	The Y coordinate of the arrow's base.
				type:	int
			ex:
				desc:	The X coordinate of the arrow's tip.
				type:	int
			ey:
				desc:	The Y coordinate of the arrow's tip..
				type:	int

		keywords:
			body_length:
				desc:	Proportional length of the arrow body relative to
						the full arrow [0-1].
				type:	float
			body_width:
				desc:	Proportional width (thickness) of the arrow body
						relative to the full arrow [0-1].
				type:	float
			head_width:
				desc:	Width (thickness) of the arrow head in pixels.
				type:	float

		Example: |
			my_canvas = Canvas()
			w = var.width/2
			h = var.height/2
			# Function interface
			my_canvas.arrow(0, 0, w, h, head_width=100, body_length=0.5)
			# Element interface
			my_canvas['my_arrow'] = Arrow(0, 0, w, h, head_width=100, body_length=0.5)
		"""

		self += Arrow(sx, sy, ex, ey, body_width=body_width,
			body_length=body_length, head_width=head_width, **style_args)
		return 'stim%d' % self._stimnr

	def rect(self, x, y, w, h, **style_args):

		"""
		desc:
			Draws a rectangle.

		arguments:
			x:
				desc:	The left X coordinate.
				type:	int
			y:
				desc:	The top Y coordinate.
				type:	int
			w:
				desc:	The width.
				type:	int
			h:
				desc:	The height.
				type:	int

		keyword-dict:
			style_args:	"%arg_style"

		example: |
			my_canvas = Canvas()
			# Function interface
			my_canvas.rect(-10, -10, 20, 20, fill=True)
			# Element interface
			my_canvas['my_rect'] = Rect(-10, -10, 20, 20, fill=True)
		"""

		self += Rect(x, y, w, h, **style_args)
		return 'stim%d' % self._stimnr

	def ellipse(self, x, y, w, h, **style_args):

		"""
		desc:
			Draws an ellipse.

		arguments:
			x:
				desc:	The left X coordinate.
				type:	int
			y:
				desc:	The top Y coordinate.
				type:	int
			w:
				desc:	The width.
				type:	int
			h:
				desc:	The height.
				type:	int

		keyword-dict:
			style_args:	"%arg_style"

		example: |
			my_canvas = Canvas()
			# Function interface
			my_canvas.ellipse(-10, -10, 20, 20, fill=True)
			# Element interface
			my_canvas['my_ellipse'] = Ellipse(-10, -10, 20, 20, fill=True)
		"""

		self += Ellipse(x, y, w, h, **style_args)
		return 'stim%d' % self._stimnr

	def polygon(self, vertices, **style_args):

		"""
		desc:
			Draws a polygon that defined by a list of vertices. I.e. a shape of
			points connected by lines.

		arguments:
			vertices:
				desc:	A list of tuples, where each tuple corresponds to a
						vertex. For example, [(100,100), (200,100), (100,200)]
						will draw a triangle.
				type:	list

		keyword-dict:
			style_args:	"%arg_style"

		example: |
			my_canvas = Canvas()
			n1 = 0,0
			n2 = 100, 100
			n3 = 0, 100
			# Function interface
			my_canvas.polygon([n1, n2, n3])
			# Element interface
			my_canvas['my_polygon'] = Polygon([n1, n2, n3])
		"""

		self += Polygon(vertices, **style_args)
		return 'stim%d' % self._stimnr

	def text_size(self, text, center=True, max_width=None, **style_args):

		"""
		desc:
			Determines the size of a text string in pixels.

		arguments:
			text:
				desc:	A string of text.
				type:	[str, unicode]

		keywords:
			center:
				desc:	"%arg_center"
				type:	bool
			max_width:
				desc:	"%arg_max_width"
				type:	[int, NoneType]

		keyword-dict:
			style_args:	"%arg_style"

		returns:
			desc:	A (width, height) tuple containing the dimensions of the
					text string.
			type:	tuple

		example: |
			my_canvas = Canvas()
			w, h = my_canvas.text_size('Some text')
		"""

		return RichText(
			text,
			center=center,
			max_width=max_width,
			font_size=self.font_size,
			font_family=self.font_family,
			font_bold=self.font_bold,
			font_italic=self.font_italic
		).construct(self.__class__(self.experiment)).size

	def text(self, text, center=True, x=None, y=None, max_width=None,
		**style_args):

		"""
		desc:
			Draws text.

		arguments:
			text:
				desc:	A string of text. When using Python 2, this should be
						either `unicode` or a utf-8-encoded `str`. When using
						Python 3, this should be either `str` or a utf-8-encoded
						`bytes`.
				type:	[str, unicode]

		keywords:
			center:
				desc:	"%arg_center"
				type:	bool
			x:
				desc:	The X coordinate, or None to draw horizontally centered
						text.
				type:	[int, NoneType]
			y:
				desc:	The Y coordinate, or None to draw vertically centered
						text.
				type:	[int, NoneType]
			max_width:
				desc:	"%arg_max_width"
				type:	[int, NoneType]

		keyword-dict:
			style_args:	"%arg_style"

		example: |
			my_canvas = Canvas()
			# Function interface
			my_canvas.text('Some text with <b>boldface</b> and <i>italics</i>')
			# Element interface
			my_canvas['my_text'] = Text('Some text with <b>boldface</b> and <i>italics</i>')
		"""

		self += RichText(text, center=center, x=x, y=y, max_width=max_width,
			**style_args)
		return 'stim%d' % self._stimnr

	def image(self, fname, center=True, x=None, y=None, scale=None,
		rotation=None):

		"""
		desc:
			Draws an image from file. This function does not look in the file
			pool, but takes an absolute path.

		arguments:
			fname:
				desc:	The filename of the image. When using Python 2, this
						should be either `unicode` or a utf-8-encoded `str`.
						When using Python 3, this should be either `str` or a
						utf-8-encoded `bytes`.
				type:	[str, unicode]

		keywords:
			center:
				desc:	A bool indicating whether coordinates indicate the
						center (True) or top-left (False).
				type:	bool
			x:
				desc:	The X coordinate, or `None` to draw a horizontally
						centered image.
				type:	[int, NoneType]
			y:
				desc:	The Y coordinate, or `None` to draw a vertically
						centered image.
				type:	[int, NoneType]
			scale:
				desc:	The scaling factor of the image. `None` or 1 indicate
						the original size. 2.0 indicates a 200% zoom, etc.
				type:	[float, int, NoneType]
			rotation:
				desc:	The rotation of the image `None` or 0 indicate the
						original rotation. Positive values indicate a
						clockwise rotation in degrees.
				type:	[float, int, NoneType]

		example: |
			my_canvas = Canvas()
			# Determine the absolute path:
			path = exp.pool[u'image_in_pool.png']
			# Function interface
			my_canvas.image(path)
			# Element interface
			my_canvas['my_image'] = Image(path)
		"""

		self += Image(
			fname, center=center, x=x, y=y, scale=scale, rotation=rotation
		)
		return 'stim%d' % self._stimnr

	def gabor(self, x, y, orient, freq, env=u'gaussian', size=96, stdev=12,
		phase=0, col1=u'white', col2=u'black', bgmode=u'avg'):

		"""
		desc: |
			Draws a Gabor patch. Note: The exact rendering of the Gabor patch
			depends on the back-end.

		arguments:
			x:
				desc:	The center X coordinate.
				type:	int
			y:
				desc:	The center Y coordinate.
				type:	int
			orient:
				desc:	|
						Orientation in degrees [0 .. 360]. This refers to a
						clockwise rotation from a vertical.

						__Version note:__ In version 3.2.6 and earlier, the
						orientation was *counterclockwise* for the *legacy* and
						*xpyriment* backends, and clockwise for the *psycho*
						backends. As of 3.2.7, the orientation is *clockwise*
						for all backends.
				type:	[float, int]
			freq:
				desc:	Frequency in cycles/px of the sinusoid.
				type:	[float, int]

		keywords:
			env:
				desc:	The envelope that determines the shape of the patch. Can
						be "gaussian", "linear", "circular", or "rectangular".
				type:	[str, unicode]
			size:
				desc:	A size in pixels.
				type:	[float, int]
			stdev:
				desc:	Standard deviation in pixels of the gaussian. Only
						applicable to gaussian envelopes.
				type:	[float, int]
			phase:
				desc:	Phase of the sinusoid [0.0 .. 1.0].
				type:	[float, int]
			col1:
				desc:	A color for the peaks.
				type:	[str, unicode]
			col2:
				desc: |
						A color for the troughs. Note: The psycho back-end
						ignores this parameter and always uses the inverse of
						`col1` for the throughs.
				type:	[str, unicode]
			bgmode:
				desc:	"%arg_bgmode"
				type:	[str, unicode]

		example: |
			my_canvas = Canvas()
			# Function interface
			my_canvas.gabor(100, 100, 45, .05)
			# Element interface
			my_canvas['my_gabor'] = Gabor(100, 100, 45, .05)
		"""

		self += Gabor(x, y, orient, freq, env=env, size=size,
			stdev=stdev, phase=phase, col1=col1, col2=col2, bgmode=bgmode)
		return 'stim%d' % self._stimnr

	def noise_patch(self, x, y, env=u'gaussian', size=96, stdev=12,
		col1=u'white', col2=u'black', bgmode=u'avg'):

		"""
		desc:
			Draws a patch of noise, with an envelope. The exact rendering of the
			noise patch depends on the back-end.

		arguments:
			x:
				desc:	The center X coordinate.
				type:	int
			y:
				desc:	The center Y coordinate.
				type:	int

		keywords:
			env:
				desc:	The envelope that determines the shape of the patch. Can
						be "gaussian", "linear", "circular", or "rectangular".
				type:	[str, unicode]
			size:
				desc:	A size in pixels.
				type:	[float, int]
			stdev:
				desc:	Standard deviation in pixels of the gaussian. Only
						applicable to gaussian envelopes.
				type:	[float, int]
			col1:
				desc:	The first color.
				type:	[str, unicode]
			col2:
				desc: |
						The second color. Note: The psycho back-end ignores this
						parameter and always uses the inverse of `col1`.
				type:	[str, unicode]
			bgmode:
				desc: 	"%arg_bgmode"
				type:	[str, unicode]

		example: |
			my_canvas = Canvas()
			# Function interface
			my_canvas.noise_patch(100, 100, env='circular')
			# Element interface
			my_canvas['my_noise_patch'] = NoisePatch(100, 100, env='circular')
		"""

		self += NoisePatch(x, y, env=env, size=size, stdev=stdev,
			col1=col1, col2=col2, bgmode=bgmode)
		return 'stim%d' % self._stimnr

	# Deprecated functions

	def xcenter(self):

		"""
		visible:	False
		desc:		deprecated
		"""

		return self.experiment.var.get(u'width') / 2

	def ycenter(self):

		"""
		visible:	False
		desc:		deprecated
		"""

		return self.experiment.var.get(u'height') / 2

	def set_penwidth(self, penwidth):

		"""
		visible:	False
		desc:		deprecated
		"""

		self.penwidth = penwidth

	def set_fgcolor(self, color):

		"""
		visible:	False
		desc:		deprecated
		"""

		self.color = color

	def set_bgcolor(self, color):

		"""
		visible:	False
		desc:		deprecated
		"""

		self.background_color = color

	def set_font(self, style=None, size=None, italic=None, bold=None,
		underline=None):

		"""
		visible:	False
		desc:		deprecated
		"""

		if style is not None: self.font_family = style
		if size is not None: self.font_size = size
		if italic is not None: self.font_italic = italic
		if bold is not None: self.font_bold = bold
		if underline is not None: self.font_underline = underline

	@staticmethod
	def init_display(experiment):

		"""
		visible: False

		desc:
			Initializes the display before the experiment begins.

		arguments:
			experiment:
				desc:	An experiment object.
				type:	experiment
		"""

		raise NotImplementedError()

	@staticmethod
	def close_display(experiment):

		"""
		visible: False

		desc:
			Closes the display after the experiment is finished.

		arguments:
			experiment:
				desc:	An experiment object.
				type:	experiment
		"""

		raise NotImplementedError()


# Translation mapping from envelope names
ENV_SYNONYMS = {
	u"c": u"c",
	u"circular": u"c",
	u"round": u"c",
	u"g": u"g",
	u"gaussian": u"g",
	u"gauss": u"g",
	u"normal": u"g",
	u"r": u"r",
	u"rectangular": u"r",
	u"rectangle": u"r",
	u"g": u"g",
	u"rect": u"g",
	u"square": u"g",
	None: u"g",
	u"l": u"l",
	u"linear": u"l",
	u"lin": u"l",
	u"ln": u"l",
	u"l": u"l"
}

canvas_cache = {}


def _color(col):

	"""
	desc:
		Wrapper function for _gabor and _noise_patch to convert color names to
		PyGame color objects.

	argumens:
		col:	A color specification.

	returns:
		A PyGame color object.
	"""

	from openexp._color.legacy import Legacy
	return Legacy(None, col).backend_color


def _xyr(x, y, size, orient=0):

	"""
	desc:
		Gets the coordinates relative to the center and compensating for the
		rotation.
	"""

	dx = x - 0.5 * size
	dy = y - 0.5 * size
	t = math.atan2(dy, dx) + orient
	r = math.sqrt(dx ** 2 + dy ** 2)
	return r * math.cos(t), r * math.sin(t), r


def _gabor(
	orient,
	freq,
	env=u"gaussian",
	size=96,
	stdev=12,
	phase=0,
	col1=u"white",
	col2=u"black",
	bgmode=u"avg"
):

	"""
	desc:
		Returns a pygame surface containing a Gabor patch. For arguments,
		see [canvas.gabor].
	"""

	env = _match_env(env)
	# Generating a Gabor patch takes quite some time, so keep
	# a cache of previously generated Gabor patches to speed up
	# the process.
	global canvas_cache
	key = u"gabor_%s_%s_%s_%s_%s_%s_%s_%s_%s" % (
		orient,
		freq,
		env,
		size,
		stdev,
		phase,
		col1,
		col2,
		bgmode
	)
	if key in canvas_cache:
		return canvas_cache[key]
	im = PILImage.new(u'RGB', (size, size))
	px = im.load()
	# Conver the orientation to radians
	orient = math.radians(-orient)
	col1 = _color(col1)
	col2 = _color(col2)
	# rx and ry reflect the real coordinates in the
	# target image
	for rx, ry in itertools.product(range(size), range(size)):
		ux, uy, r = _xyr(rx, ry, size, orient)
		# Get the amplitude without the envelope (0 .. 1)
		amp = 0.5 + 0.5 * math.cos(2.0 * math.pi * (ux * freq + phase))
		# The envelope adjustment
		if env == "g":
			f = math.exp(-0.5 * (ux / stdev) ** 2 - 0.5 * (uy / stdev) ** 2)
		elif env == "l":
			f = max(0, (0.5 * size - r) / (0.5 * size))
		elif env == "c":
			if (r > 0.5 * size):
				f = 0.0
			else:
				f = 1.0
		else:
			f = 1.0
		# Apply the envelope
		if bgmode == u"avg":
			amp = amp * f + 0.5 * (1.0 - f)
		elif bgmode == u"col2":
			amp = amp * f
		else:
			raise osexception(
				u"Invalid argument for bgmode: %s "
				u"(should be one of 'avg','col2')"
				% bgmode
			)
		r = col1.r * amp + col2.r * (1.0 - amp)
		g = col1.g * amp + col2.g * (1.0 - amp)
		b = col1.b * amp + col2.b * (1.0 - amp)
		px[rx, ry] = int(r), int(g), int(b)
	canvas_cache[key] = im
	return im


def _noise_patch(
	env=u"gaussian",
	size=96,
	stdev=12,
	col1=u"white",
	col2=u"black",
	bgmode=u"avg"
):

	"""
	desc:
		Returns a pygame surface containing a noise patch. For arguments,
		see [canvas.noise_patch].
	"""

	env = _match_env(env)
	# Generating a noise patch takes quite some time, so keep
	# a cache of previously generated noise patches to speed up
	# the process.
	global canvas_cache
	key = u"noise_%s_%s_%s_%s_%s_%s" % (env, size, stdev, col1, col2, bgmode)
	if key in canvas_cache:
		return canvas_cache[key]
	# Create a surface
	im = PILImage.new(u'RGB', (size, size))
	px = im.load()
	col1 = _color(col1)
	col2 = _color(col2)
	# rx and ry reflect the real coordinates in the
	# target image
	for rx, ry in itertools.product(range(size), range(size)):
		ux, uy, r = _xyr(rx, ry, size)
		# Get the amplitude without the envelope (0 .. 1)
		amp = random.random()
		# The envelope adjustment
		if env == u"g":
			f = math.exp(-0.5 * (ux / stdev) ** 2 - 0.5 * (uy / stdev) ** 2)
		elif env == u"l":
			f = max(0, (0.5 * size - r) / (0.5 * size))
		elif env == u"c":
			if (r > 0.5 * size):
				f = 0.0
			else:
				f = 1.0
		else:
			f = 1.0
		# Apply the envelope
		if bgmode == u"avg":
			amp = amp * f + 0.5 * (1.0 - f)
		elif bgmode == u"col2":
			amp = amp * f
		else:
			raise osexception(
				u"Invalid argument for bgmode: %s "
				u"(should be one of 'avg','col2')"
				% bgmode
			)
		r = col1.r * amp + col2.r * (1.0 - amp)
		g = col1.g * amp + col2.g * (1.0 - amp)
		b = col1.b * amp + col2.b * (1.0 - amp)
		px[rx, ry] = int(r), int(g), int(b)
	canvas_cache[key] = im
	return im


def _match_env(env):

	"""
	desc:
		Translation between various envelope names.

	arguments:
		env:
			desc:	An envelope name.
			type:	[str, unicode]

	returns:
		desc: A standard envelope name ("c", "g", "r" or "l")
		type: unicode
	"""

	if env not in ENV_SYNONYMS:
		raise osexception(u"'%s' is not a valid envelope" % env)
	return ENV_SYNONYMS[env]


# Non PEP-8 alias for backwards compatibility
canvas = Canvas



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtCore, QtGui, QtWidgets

class variable_inspector_cell(QtWidgets.QTableWidgetItem):

	"""
	desc:
		A cell for the variable-inspector table.
	"""

	def __init__(self, text, info):

		"""
		desc:
			Constructor.

		arguments:
			text:
				desc:	The cell text.
				type:	str
			info:
				desc:	Variable info as returned by var_store.inspect().
				type:	dict
		"""

		a = QtCore.Qt.AlignLeft
		f = QtGui.QFont()
		if info[u'alive']:
			f.setBold(True)
		if not isinstance(text, basestring):
			try:
				len(text)
			except TypeError:
				if text is None:
					text = u''
				elif isinstance(text, int) or isinstance(text, float):
					text = str(text)
					a = QtCore.Qt.AlignRight
				else:
					text = safe_decode(text, errors=u'ignore')
			else:
				text = u','.join([safe_decode(s) for s in text])
		QtWidgets.QTableWidgetItem.__init__(self, text)
		self.setFont(f)
		self.setTextAlignment(a)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame import item
from libqtopensesame import qtplugin

from qtpy import QtWidgets, QtCore

import warnings
import os

import imp

try:
	if os.name == 'posix':
		# import the local modified version of pyparallel
		# that allows for non-exclusive connections to the parport
		path_to_file = os.path.join(os.path.dirname(__file__), "parallelppdev.py")
		parallel = imp.load_source('parallel', path_to_file)
		#import parallelppdev as parallel
	else:
		import parallel
except ImportError:
	warnings.warn("The parallel module could not load, please make sure you have installed pyparallel.")

# we only want one instance of pp, so here's a global var
_pp = None

class pp_io(item.item):

	"""
	This class (the class with the same name as the module)
	handles the basic functionality of the item. It does
	not deal with GUI stuff.
	"""

	def __init__(self, name, experiment, string = None):

		"""
		Constructor
		"""

		# The item_typeshould match the name of the module
		self.item_type = "pp_io"

		# Provide a short accurate description of the item's functionality
		self.description = "Allows setting pins on the parallel port"

		# Set some item-specific variables
		self.value = 1
		self.duration = 1

		# The parent handles the rest of the contruction
		item.item.__init__(self, name, experiment, string)

		# add cleanup code
		self.experiment.cleanup_functions.append(self.clean_up_the_mess)

	def clean_up_the_mess(self):
		global _pp
		if self.experiment.debug:
			print("pp_io.clean_up_this_mess(): deleting _pp")
		if not _pp is None:
			del _pp

	def prepare(self):

		"""
		Prepare the item. In this case this means doing little.
		"""

		# Pass the word on to the parent
		item.item.prepare(self)

		# get the global pp instance and initialize it if
		# necessary
		global _pp
		if _pp is None:
			try:
				_pp = parallel.Parallel()
			except OSError:
				warnings.warn("Could not access /dev/parport0.")
		self.pp = _pp

		# This function prepares a self._duration_func() function based
		# on the compensation and duration variables.
		self.prepare_duration()

		# Report success
		return True

	def run(self):

		"""
		Run the item. In this case this means putting the offline canvas
		to the display and waiting for the specified duration.
		"""

		# Set the pp value
		if not self.pp is None:
			self.set_item_onset(self.pp.setData(self.value))

		# This function has been prepared by self.prepare_duration()
		self._duration_func()

		# unless duration was zero, turn it off
		if not self.pp is None and self.duration !=0:
			self.pp.setData(0)

		# Report success
		return True

class qtpp_io(pp_io, qtplugin.qtplugin):

	"""
	This class (the class named qt[name of module] handles
	the GUI part of the plugin. For more information about
	GUI programming using qtpy, see:
	<http://www.riverbankcomputing.co.uk/static/Docs/qtpy/html/classes.html>
	"""

	def __init__(self, name, experiment, string = None):

		"""
		Constructor
		"""

		# Pass the word on to the parents
		pp_io.__init__(self, name, experiment, string)
		qtplugin.qtplugin.__init__(self, __file__)

	def init_edit_widget(self):

		"""
		This function creates the controls for the edit
		widget.
		"""

		# Lock the widget until we're doing creating it
		self.lock = True

		# Pass the word on to the parent
		qtplugin.qtplugin.init_edit_widget(self, False)

		# Create the controls
		#
		# A number of convenience functions are available which
		# automatically create controls, which are also automatically
		# updated and applied. If you set the varname to None, the
		# controls will be created, but not automatically updated
		# and applied.
		#
		# qtplugin.add_combobox_control(varname, label, list_of_options)
		# - creates a QComboBox
		# qtplugin.add_line_edit_control(varname, label)
		# - creates a QLineEdit
		# qtplugin.add_spinbox_control(varname, label, min, max, suffix = suffix, prefix = prefix)

		self.add_spinbox_control("value", "Value", 0, 255, tooltip = "Value to set port")
		self.add_line_edit_control("duration", "Duration", tooltip = "Expecting a value in milliseconds, 'keypress' or 'mouseclick'")

		# Add a stretch to the edit_vbox, so that the controls do not
		# stretch to the bottom of the window.
		self.edit_vbox.addStretch()

		# Unlock
		self.lock = True

	def apply_edit_changes(self):

		"""
		Set the variables based on the controls
		"""

		# Abort if the parent reports failure of if the controls are locked
		if not qtplugin.qtplugin.apply_edit_changes(self) or self.lock:
			return False

		# Refresh the main window, so that changes become visible everywhere
		self.experiment.main_window.refresh(self.name)

		# Report success
		return True

	def edit_widget(self):

		"""
		Set the controls based on the variables
		"""

		# Lock the controls, otherwise a recursive loop might aris
		# in which updating the controls causes the variables to be
		# updated, which causes the controls to be updated, etc...
		self.lock = True

		# Let the parent handle everything
		qtplugin.qtplugin.edit_widget(self)

		# Unlock
		self.lock = False

		# Return the _edit_widget
		return self._edit_widget



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame import misc
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from libqtopensesame.widgets.base_widget import base_widget
from libqtopensesame._input.popup_menu import popup_menu
from libqtopensesame._input.confirmation import confirmation
from qtpy import QtCore, QtWidgets
import os
import os.path
import shutil
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'pool_widget', category=u'core')


class pool_widget(base_widget):

	"""
	desc:
		The file-pool widget.
	"""

	def __init__(self, main_window):

		"""
		desc:
			Constructor.

		arguments:
			main_window: The main-window object.
		"""

		self.max_len = 5
		super(pool_widget, self).__init__(main_window,
			ui=u'widgets.pool_widget')
		self.ui.button_pool_add.clicked.connect(self.select_and_add)
		self.ui.button_refresh.clicked.connect(self.refresh)
		self.ui.button_help_pool.clicked.connect(self.help)
		self.ui.button_browse_pool.clicked.connect(self.browse)
		self.ui.edit_pool_filter.textChanged.connect(self.refresh)
		self.ui.combobox_view.currentIndexChanged.connect(self.set_view)
		self.ui.list_pool.itemActivated.connect(self.activate_file)
		self.ui.list_pool.itemChanged.connect(self.rename_file)
		self.main_window.theme.apply_theme(self)
		self.ui.label_size_warning.setVisible(False)

	def help(self):

		"""
		desc:
			Opens the help tab.
		"""

		self.tabwidget.open_osdoc(u'manual/interface')

	def set_view(self):

		"""
		desc:
			Sets the viewmode (list/ icons) based on the gui control.
		"""

		if self.ui.combobox_view.currentIndex() == 0:
			self.ui.list_pool.setViewMode(QtWidgets.QListView.ListMode)
		else:
			self.ui.list_pool.setViewMode(QtWidgets.QListView.IconMode)

	def browse(self):

		"""
		desc:
			Opens the pool folder in the file manager in an OS specific way.
		"""

		misc.open_url(self.experiment.pool.folder())

	def add(self, files, rename=False):

		"""
		desc:
			Adds a list of files to the pool.

		arguments:
			files:
			 	desc:	A list of paths.
				type:	list
		"""

		if not files:
			return
		for path in files:
			basename = os.path.basename(path)
			if self.pool.in_folder(basename):
				if rename:
					while self.pool.in_folder(basename):
						basename = u'_' + basename
				else:
					c = confirmation(self.main_window,
						_(u"A file named '%s' already exists in the pool. Do you want to overwrite this file?") \
						% basename)
					if not c.show():
						continue
			try:
				self.pool.add(path, new_name=basename)
			except (IOError, shutil.Error) as e:
				self.notify(_(u'Failed to copy %s to file pool') % path)
				self.console.write(safe_decode(e, errors=u'ignore'))
		self.refresh()
		self.select(basename)

	def select_and_add(self, dummy=None):

		"""
		desc:
			Adds one or more files to the pool.

		"""

		path_list = QtWidgets.QFileDialog.getOpenFileNames(
			self.main_window.ui.centralwidget, _(u"Add files to pool"),
			directory=cfg.default_pool_folder)
		if not path_list:
			return
		# Qt5 puts the first element in another list, check for that here
		if isinstance(path_list[0], list):
			path_list = path_list[0]
		if not path_list:
			return
		cfg.default_pool_folder = os.path.dirname(path_list[0])
		self.add(path_list)

	def select(self, fname):

		"""
		desc:
			Selects a specific file in the file pool.

		arguments:
			fname:
				desc:	The file to be selected.
				type:	[str, unicode]
		"""

		for i in range(self.ui.list_pool.count()):
			item = self.ui.list_pool.item(i)
			if item.text() == fname:
				self.ui.list_pool.setCurrentItem(item)

	def refresh(self):

		"""
		desc:
			Refreshes the contents of the pool widget.
		"""

		try:
			path_iterator = iter(self.pool)
		except Exception as e:
			self.notify(_(u'Failed to refresh file pool'))
			self.console.write(safe_decode(e, errors=u'ignore'))
			return
		filt = self.ui.edit_pool_filter.text().lower()
		self.ui.list_pool.clear()
		for path in path_iterator:
			oslogger.debug(path)
			fname = os.path.basename(path)
			if filt in fname.lower():
				icon = self.theme.qfileicon(self.pool[path])
				item = QtWidgets.QListWidgetItem(icon, fname)
				item.setFlags(item.flags()|QtCore.Qt.ItemIsEditable)
				item.icon = icon
				item.path = path
				item.setToolTip(path)
				self.ui.list_pool.addItem(item)
		try:
			size = self.pool.size()
		except:
			size = -1
		oslogger.debug(u'pool is %d bytes' % size)
		if size > cfg.file_pool_size_warning:
			self.ui.label_size_warning.setText(_('Your file pool is larger '
				'than usual (%d MB). This increases loading and saving time. '
				'Consider moving files from the file pool to the experiment '
				'folder.') % (size / 1048576))
			self.ui.label_size_warning.setVisible(True)
		else:
			self.ui.label_size_warning.setVisible(False)

	def open_file(self, path):

		"""
		desc:
			Opens a file in a platform specific way.

		arguments:
			path:
				desc:	The file to be opened.
				type:	[unicode, str]
		"""

		misc.open_url(self.pool[path])

	def activate_file(self, item):

		"""
		desc:
			Is called when a file is double-clicked or otherwise activated and
			opens the file.

		arguments:
			item:
				type:	QListWidgetItem
		"""

		self.open_file(item.path)

	def contextMenuEvent(self, event):

		"""
		desc:
			Presents a context menu.

		arguments:
			event:
			 	type:	QMouseClickEvent
		"""

		item = self.ui.list_pool.itemAt(self.ui.list_pool.mapFromGlobal(
			event.globalPos()))
		if item is None:
			return
		path = item.path
		pm = popup_menu(self, [
			(0, _(u'Open'), item.icon),
			(1, _(u'Remove from pool'), u'delete'),
			(2, _(u'Rename'), u'rename'),
			])
		action = pm.show()
		if action == 0:
			self.open_file(path)
		elif action == 1:
			self.delete_selected_files()
		elif action == 2:
			self.ui.list_pool.editItem(item)

	def delete_selected_files(self):

		"""
		desc:
			Deletes all selected files from the pool. Asks for confimation\
			first.
		"""

		# Prepare the confirmation dialog, which contains a limited nr of
		# filenames
		l = []
		suffix = u''
		for item in self.ui.list_pool.selectedItems()[:self.max_len]:
			l.append(self.pool[item.text()])
		if len(self.ui.list_pool.selectedItems()) > self.max_len:
			suffix = _('And %d more file(s)') % \
				(len(self.ui.list_pool.selectedItems())-self.max_len)
		msg = _(u"<p>Are you sure you want to remove the following files?</p>"
			u"<p>- %s</p> <p>%s</p>") % (u"<br /> - ".join(l), suffix)
		c = confirmation(self.main_window, msg)
		if not c.show():
			return
		# Create a list of files to be removed
		dL = []
		for item in self.ui.list_pool.selectedItems():
			dL.append(item.path)
		# Remove the files
		try:
			for f in dL:
				del self.pool[f]
			oslogger.debug(u"removed '%s'" % f)
		except:
			oslogger.error(u"failed to remove '%s'" % f)
		self.refresh()
		self.main_window.set_unsaved()

	def rename_file(self, item):

		"""
		desc:
			Starts a rename action for an item.

		arguments:
			item:
				type:	QListWidgetItem
		"""

		old_name = item.path
		new_name = item.text().strip()
		if new_name in self.pool:
			self.notify(
				_(u"There already is a file named '%s' in the file pool") \
				% new_name)
			new_name = old_name
		elif new_name in (old_name, u''):
			new_name = old_name
		else:
			try:
				self.pool.rename(old_name, new_name)
			except:
				self.notify(_(u'Failed to rename "%s" to "%s".') \
					% (old_name, new_name))
		self.refresh()
		self.select(new_name)

	def dragEnterEvent(self, event):

		"""
		desc:
			Accepts an incoming drag that precedes a drop.

		arguments:
			event:
			 	type:	QDragEnterEvent
		"""

		event.acceptProposedAction()

	def dropEvent(self, event):

		"""
		desc:
			Accepts an incoming drop.

		arguments:
			event:
			 	type:	QDropEnterEvent
		"""
		files = []
		for url in event.mimeData().urls():
			files.append(url.toLocalFile())
		self.add(files)
		event.acceptProposedAction()

	def focusInEvent(self, e):

		"""
		desc:
			Focus the filter edit when the widget receives focus.

		arguments:
			e:
				type:	QFocusEvent
		"""

		self.ui.edit_pool_filter.setFocus()

def select_from_pool(main_window, parent=None):

	"""
	desc:
		A static function that presents the select from pool dialog.

	arguments:
		main_window:	The GUI main window

	keywords:
		parent:		The parent QWidget or None to use main window's central
					widget.
	"""

	if parent is None:
		parent = main_window.ui.centralwidget
	d = QtWidgets.QDialog(parent)
	widget = pool_widget(main_window)
	widget.refresh()
	bbox = QtWidgets.QDialogButtonBox(d)
	bbox.addButton(_(u"Cancel"), QtWidgets.QDialogButtonBox.RejectRole)
	bbox.addButton(_(u"Select"), QtWidgets.QDialogButtonBox.AcceptRole)
	bbox.accepted.connect(d.accept)
	bbox.rejected.connect(d.reject)
	vbox = QtWidgets.QVBoxLayout()
	vbox.addWidget(widget)
	vbox.addWidget(bbox)
	d.setLayout(vbox)
	d.setWindowTitle(_(u"Select file from pool"))
	res = d.exec_()
	main_window.ui.pool_widget.refresh()
	if res == QtWidgets.QDialog.Rejected:
		return u""
	selected = widget.ui.list_pool.currentItem()
	if selected is None:
		return u""
	return selected.text()



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import pygame
from libopensesame.exceptions import osexception
from openexp._canvas._image.image import Image
from openexp._canvas._element.legacy import LegacyElement


class Legacy(LegacyElement, Image):

	def prepare(self):

		if not hasattr(self, '_image_surface') or self._dirty:
			self._dirty = False
			fname = safe_decode(self.fname)
			if not os.path.isfile(fname):
				raise osexception(u'"%s" does not exist' % fname)
			with open(fname, u'rb') as fd:
				try:
					self._image_surface = pygame.image.load(fd)
				except pygame.error:
					raise osexception(
						u"'%s' is not a supported image format" % fname)
			# After rotation, the figure gets bigger. We therefore need to
			# compensate by moving it a bit
			if self.rotation is not None and self.rotation != 0:
				w1, h1 = self._image_surface.get_size()
				self._image_surface = pygame.transform.rotate(
					self._image_surface.convert_alpha(), -self.rotation
				)
				w2, h2 = self._image_surface.get_size()
				self._dx = (w2-w1)/2
				self._dy = (h2-h1)/2
			else:
				self._dx = self._dy = 0
			if self.scale is not None:
				try:
					self._image_surface = pygame.transform.smoothscale(
						self._image_surfacesurface,
						(int(self._image_surface.get_width()*self.scale),
						int(self._image_surface.get_height()*self.scale))
					)
				except:
					self._image_surface = pygame.transform.scale(
						self._image_surface,
						(int(self._image_surface.get_width()*self.scale),
						int(self._image_surface.get_height()*self.scale))
					)
				self._dx *= self.scale
				self._dy *= self.scale
		size = self._image_surface.get_size()
		x, y = self.to_xy(self.x, self.y)
		if self.center:
			x -= size[0] / 2
			y -= size[1] / 2
		else:
			x -= self._dx
			y -= self._dy
		self.surface.blit(self._image_surface, (x, y))

	@staticmethod
	def _setter(key, self, val):

		self._dirty = True
		Image._setter(key, self, val)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

__author__ = "Sebastiaan Mathot"
__license__ = "GPLv3"

from qtpy import QtWidgets
import os.path

class recent_action(QtWidgets.QAction):

	"""Menu action for a recently opened file"""

	def __init__(self, path, main_window, menu):

		"""
		Constructor

		Arguments:
		path -- path to the recent file
		main_window -- the main window
		menu -- the menu into which the action should be inserted
		"""

		QtWidgets.QAction.__init__(self, os.path.basename(path), menu)
		self.main_window = main_window
		self.triggered.connect(self.open_file)
		self.path = path

	def open_file(self, dummy = None):

		"""
		Open the file

		Keyword arguments:
		dummy -- a dummy argument passed by the signal handler (default=None)
		"""

		self.main_window.open_file(path=self.path)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._canvas._ellipse.ellipse import Ellipse
from openexp._canvas._element.legacy import LegacyElement
import pygame


class Legacy(LegacyElement, Ellipse):

	def prepare(self):

		x = int(self.x)
		y = int(self.y)
		w = int(self.w)
		h = int(self.h)
		x, y = self.to_xy(x, y)
		if self.fill:
			pygame.draw.ellipse(self.surface, self.color.backend_color,
				(x, y, w, h), 0)
			return
		# Use experyiment's method to draw ellipses with a transparent
		# interior by using the transparent colorkey method. This involves
		# some trickery by making a separate surface of which a certain color
		# is designated as transparent and draw the ellipse on there first.
		# When being blit onto another surface, this transparent color
		# (e.g. the ellipses interior is not blitted with the rest.
		line_width = self.penwidth
		size = (w, h)
		surface = pygame.surface.Surface(
			[p + line_width for p in size],
			pygame.SRCALPHA).convert_alpha()
		pygame.draw.ellipse(surface, (0, 0, 0), pygame.Rect(
			(0, 0),
			[p + line_width for p in size]))
		tmp = pygame.surface.Surface(
			[p - line_width for p in size])
		tmp.fill([0, 0, 0])
		tmp.set_colorkey([255, 255, 255])
		hole = pygame.surface.Surface(
			[p - line_width for p in size],
			pygame.SRCALPHA).convert_alpha()
		pygame.draw.ellipse(tmp, (255, 255, 255), pygame.Rect(
			(0, 0), [p - line_width for p in size]))
		hole.blit(tmp, (0, 0))
		surface.blit(hole, (line_width, line_width),
			special_flags=pygame.BLEND_RGBA_MIN)
		surface.fill(self.color.backend_color,
			special_flags=pygame.BLEND_RGB_MAX)
		# The line_width affects the temp's surface size, so use it to correct the
		# positioning when blitting.
		self.surface.blit(surface, (x-line_width/2, y-line_width/2) )



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import sys
import platform

# On Linux this appears to be buggy
if platform.system() != 'Linux':
	os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
# solves a library conflict for Linux with Nvidia drivers
# See https://forum.qt.io/topic/81328/ubuntu-qopenglshaderprogram-shader-program-is-not-linked/2
else:
	try:
		from OpenGL import GL
	except ImportError:  # Rapunzel doesn't need OpenGL
		pass

# Attach a dummy console when launched with pythonw.exe. This is necessary for
# some libraries.
if sys.executable.endswith('pythonw.exe'):
	sys.stdout = open(os.devnull, 'w')
	sys.stderr = open(os.devnull, 'w')
	sys.stdin = open(os.devnull)

# On MacOS, the locations of the SSL cerfificates need to be explicitly indicated when packaged
# as an app, otherwise urllib.urlopen fails with an SSL related error. To do so, we make use of the
# certifi package that can provide this location, but it is unclear if certifi is always installed,
# so let's fail gracefully if it isn't.
if platform.system() == 'Darwin':
	try:
		import certifi
		os.environ['SSL_CERT_FILE'] = certifi.where()
		os.environ['SSL_CERT_DIR'] = os.path.dirname(certifi.where())
	except ImportError:
		pass # Or log something here?
	# Workaround for an incompatibility between Qt and Mac OS Big Sur. See:
	# - https://forum.cogsci.nl/discussion/comment/21525/#Comment_21525
	os.environ['QT_MAC_WANTS_LAYER'] = '1'

def patch_pyqt():

	"""This patches PyQt such that properties are removed from objects before
	connectSlotsByName() tries to inspect the object. This is necessary because
	in some versions, the introspection crashes when properties cannot be
	accessed, for example because the object needs to be initialized first.
	"""

	def _(fnc):

		def make_object_safe(obj, cls):

			tmp = {}
			for name, value in list(cls.__dict__.items()):
				try:
					getattr(obj, name)
				except AttributeError:
					tmp[name] = value
					delattr(cls, name)
			return tmp

		def restore_object(cls, tmp):

			for name, value in tmp.items():
				setattr(cls, name, value)

		def inner(obj):

			tmp = {}
			for cls in obj.__class__.mro():
				tmp[cls] = make_object_safe(obj, cls)
			fnc(obj)
			for cls in tmp:
				restore_object(cls, tmp[cls])

		return inner

	from qtpy import QtCore
	QtCore.QMetaObject.connectSlotsByName = \
		_(QtCore.QMetaObject.connectSlotsByName)


def set_paths():

	from libopensesame import misc
	import sys
	import platform
	# On Windows, the working directory should be the folder with the Python
	# interpreter. If not, Qt will have difficulty finding the correct paths.
	if platform.system() == 'Windows':
		os.chdir(os.path.dirname(sys.executable))
	# Update the system path so that Qt can find itself
	path = os.path.join(os.getcwd(), 'Library', 'bin')
	if os.path.exists(path):
		os.environ['PATH'] = path + ';' + os.environ['PATH']
	# Add the Scripts subfolder, which is where Anaconda scripts are located
	path = os.path.join(os.getcwd(), 'Scripts')
	if os.path.exists(path):
		os.environ['PATH'] = path + ';' + os.environ['PATH']
	from qtpy import QtCore
	# Add the folder that contains the OpenSesame modules to the path. This is
	# generally only necessary if OpenSesame is directly run from source,
	# instead from an installation.
	if os.path.exists(os.path.join(os.getcwd(), 'libopensesame')):
		sys.path.insert(0, os.getcwd())
	# Add the Qt plugin folders to the library path, if they exists. Where
	# these folders are depends on the version of Qt4, but these are two
	# possible locations.
	qt_plugin_path = os.path.join(
		os.path.dirname(sys.executable), 'Library', 'plugins')
	if os.path.isdir(qt_plugin_path):
		QtCore.QCoreApplication.addLibraryPath(
			safe_decode(qt_plugin_path, enc=misc.filesystem_encoding())
		)
	qt_plugin_path = os.path.join(
		os.path.dirname(sys.executable), 'Library', 'lib', 'qt4', 'plugins')
	if os.path.isdir(qt_plugin_path):
		QtCore.QCoreApplication.addLibraryPath(
			safe_decode(qt_plugin_path, enc=misc.filesystem_encoding())
		)


def opensesame():

	set_paths()
	if py3:
		patch_pyqt()
	# Support for multiprocessing when packaged
	# In OS X the multiprocessing module is horribly broken, but a fixed
	# version has been released as the 'billiard' module
	if platform.system() == 'Darwin':
		# Use normal multiprocessing module from python 3.4 and on
		if sys.version_info >= (3, 4):
			from multiprocessing import freeze_support, set_start_method
			freeze_support()
			set_start_method('spawn')
		else:
			from billiard import freeze_support, forking_enable
			freeze_support()
			forking_enable(0)
	else:
		from multiprocessing import freeze_support
		freeze_support()
	# Parse the (optional) environment file that contains special paths, etc.
	from libopensesame.misc import parse_environment_file
	parse_environment_file()
	# Force the new-style Qt API
	import sip
	import qtpy
	sip.setapi('QString', 2)
	sip.setapi('QVariant', 2)
	# Do the basic window initialization
	from qtpy.QtWidgets import QApplication
	# From Qt 5.6 on, QtWebEngine is the default way to render web pages
	# QtWebEngineWidgets must be imported before a QCoreApplication instance is
	# created.
	try:
		from qtpy import QtWebEngineWidgets
	except ImportError:
		pass
	if platform.system() != 'Linux':
		# Enable High DPI pixmaps with PyQt5
		if hasattr(qtpy.QtCore.Qt, 'AA_UseHighDpiPixmaps'):
			QApplication.setAttribute(qtpy.QtCore.Qt.AA_UseHighDpiPixmaps)
		# Enable scaling for High DPI displays with PyQt5
		if hasattr(qtpy.QtCore.Qt, 'AA_EnableHighDpiScaling'):
			QApplication.setAttribute(qtpy.QtCore.Qt.AA_EnableHighDpiScaling)
	from libqtopensesame.qtopensesame import qtopensesame
	app = QApplication(sys.argv)
	opensesame = qtopensesame(app)
	opensesame.__script__ = __file__
	app.processEvents()
	# Install the translator. For some reason, the translators need to be
	# instantiated here and not in the set_locale() function, otherwise the
	# locale is ignored.
	from qtpy.QtCore import QTranslator
	opensesame.set_locale([
		QTranslator(),
		QTranslator(),
		QTranslator(),
		QTranslator()
	])
	# Now that the window is shown, load the remaining modules and resume the
	# GUI initialization.
	opensesame.resume_init()
	opensesame.restore_window_state()
	opensesame.show()
	# Added for OS X, otherwise Window will not appear
	opensesame.raise_()
	# Exit using the application exit status
	sys.exit(app.exec_())


def opensesamerun():

	set_paths()
	from libopensesame.oslogging import oslogger
	oslogger.start(u'gui')
	import libopensesame.misc
	libopensesame.misc.parse_environment_file()
	import libopensesame.experiment
	# Parse the command line options
	options = libopensesame.misc.opensesamerun_options()
	app = None
	# If the command line options haven't provided sufficient information to
	# run right away, present a GUI
	while not libopensesame.misc.opensesamerun_ready(options):
		# If PyQt4 is not available (e.g., this might be the case on Mac OS)
		# give an error instead of showing a GUI. This makes sure that even
		# without PyQt4, people can still run experiments.
		try:
			# Change Qt API
			import sip
			sip.setapi('QString', 2)
			sip.setapi('QVariant', 2)
			from qtpy import QtGui, QtCore, QtWidgets
		except:
			libopensesame.misc.messagebox(u"OpenSesame Run", u"Incorrect or "
				u"missing options.\n\nRun 'opensesame --help' from a terminal "
				u"(or command prompt) to see a list of available options, or "
				u"install Python Qt4 to enable the graphical user interface.")
			sys.exit()
		# Create the GUI and show it
		import libqtopensesame.qtopensesamerun
		if app is None:
			app = QtWidgets.QApplication(sys.argv)
			myapp = libqtopensesame.qtopensesamerun.qtopensesamerun(options)
		myapp.show()
		app.exec_()
		# Update the options from the GUI
		options = myapp.options
		# Exit if the GUI was canceled
		if not myapp.run:
			sys.exit()
	# Decode the experiment path and logfile
	experiment = os.path.abspath(options.experiment)
	if isinstance(experiment, bytes):
		experiment = safe_decode(
			experiment,
			enc=libopensesame.misc.filesystem_encoding(),
			errors=u'ignore'
		)
	# experiment_path = os.path.dirname(experiment)
	logfile = options.logfile
	if isinstance(logfile, bytes):
		logfile = safe_decode(
			logfile,
			enc=libopensesame.misc.filesystem_encoding(),
			errors=u'ignore'
		)
	experiment_path = safe_decode(
		os.path.abspath(options.experiment),
		enc=libopensesame.misc.filesystem_encoding()
	)
	if options.debug:
		# In debug mode, don't try to catch any exceptions
		exp = libopensesame.experiment.experiment(
			u"Experiment",
			experiment,
			experiment_path=experiment_path
		)
		exp.set_subject(options.subject)
		exp.var.fullscreen = options.fullscreen
		exp.logfile = logfile
		exp.run()
		exp.end()
	else:
		# Try to parse the experiment from a file
		try:
			exp = libopensesame.experiment.experiment(
				u"Experiment",
				experiment,
				experiment_path=experiment_path
			)
		except Exception as e:
			libopensesame.misc.messagebox(
				u"OpenSesame Run",
				libopensesame.misc.strip_tags(e)
			)
			sys.exit()
		# Set some options
		exp.set_subject(options.subject)
		exp.var.fullscreen = options.fullscreen
		exp.logfile = logfile
		# Initialize random number generator
		import random
		random.seed()
		# Try to run the experiment
		try:
			exp.run()
		except Exception as e:
			# Try to nicely end the experiment, even though an exception
			# occurred.
			try:
				exp.end()
			except Exception as f:
				libopensesame.misc.messagebox(
					u"OpenSesame Run",
					libopensesame.misc.strip_tags(f)
				)
			libopensesame.misc.messagebox(
				u"OpenSesame Run",
				libopensesame.misc.strip_tags(e)
			)
	libopensesame.experiment.clean_up(exp.debug)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._canvas._image.image import Image
from openexp._canvas._element.xpyriment import XpyrimentElement
from expyriment.stimuli import Picture


class Xpyriment(XpyrimentElement, Image):

	def prepare(self):

		self._stim = Picture(filename=safe_decode(self.fname))
		if self.rotation is not None and self.rotation != 0:
			w1, h1 = self._stim.surface_size
			self._stim.rotate(-self.rotation)
			w2, h2 = self._stim.surface_size
			dx = (w2-w1)/2
			dy = (h2-h1)/2
		else:
			dx = dy = 0
		if self.scale is not None:
			self._stim.scale((self.scale, self.scale))
			dx *= self.scale
			dy *= self.scale
		x, y = self.to_xy(self.x, self.y)
		if not self.center:
			w, h = self._stim.surface_size
			x += w//2-dx
			y -= h//2-dy
		self._stim.reposition((x, y))
		self._stim.preload()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
import platform
import sys
import os
import signal

if platform.system() == 'Darwin' and \
	sys.version_info < (3,4):
		# In OS X Python < 3.4 the multiprocessing module is horribly broken,
		# but a fixed version has been released as the 'billiard' module
		import billiard as multiprocessing
else:
	import multiprocessing

class OutputChannel:

	"""Passes messages from child process back to main process."""

	def __init__(self, channel, orig=None):

		"""
		Constructor.

		Arguments:
		channel	--	A multiprocessing.JoinableQueue object that is referenced
					from the main process.

		Keyword arguments:
		orig	--	The original stdout or stderr to also print the messages to.
		"""

		self.channel = channel
		self.orig = orig

	def write(self, m):

		"""
		Writes a message to the queue.

		Arguments
		m		--	The message to write. Should be a string or an (Exception,
					traceback) tuple.
		"""

		self.channel.put(m)

	def flush(self):

		"""Dummy function to mimic the stderr.flush() function."""

		if self.orig:
			self.orig.flush()
		else:
			pass

	def isatty(self):

		"""
		desc:
			Indicates that the output is not attached to a terminal.
		"""

		return False


class ExperimentProcess(multiprocessing.Process):

	"""Creates a new process to run an experiment in."""

	def __init__(self, exp, output):

		"""
		Constructor.

		Arguments
		exp		--	An instance of libopensesame.experiment.experiment
		output	--	A reference to the queue object created in and used to
					communicate with the main process.
		"""

		multiprocessing.Process.__init__(self)
		self.output = output
		# The experiment object is troublesome to serialize,
		# therefore pull out all relevant data to pass on to the new process
		# and rebuild the exp object in there.
		self.script = exp.to_string()
		self.pool_folder = exp.pool.folder()
		self.subject_nr = exp.var.subject_nr
		self.experiment_path = exp.experiment_path
		self.fullscreen = exp.var.fullscreen == u'yes'
		self.logfile = exp.logfile
		self.auto_response = exp.auto_response
		self.killed = False

	def run(self):

		"""
		Everything in this function is run in a new process, therefore all
		import statements are put in here. The function reroutes all output to
		stdin and stderr to the pipe to the main process so OpenSesame can
		handle all prints and errors.
		"""

		import os
		import sys
		from libopensesame import misc
		from libopensesame.experiment import experiment
		from libopensesame.exceptions import osexception
		# Under Windows, change the working directory to the OpenSesame folder,
		# so that the new process can find the main script.
		if os.name == u'nt':
			os.chdir(misc.opensesame_folder())
			os.environ['PATH'] += ';' + os.getcwd()
		# Reroute output to OpenSesame main process, so everything will be
		# printed in the Debug window there.
		pipeToMainProcess = OutputChannel(self.output)
		sys.stderr = sys.stdout = pipeToMainProcess
		oslogger.start(u'runtime')
		# First initialize the experiment and catch any resulting Exceptions
		try:
			exp = experiment(
				string=self.script, pool_folder=self.pool_folder,
				experiment_path=self.experiment_path,
				fullscreen=self.fullscreen, auto_response=self.auto_response,
				subject_nr=self.subject_nr, logfile=self.logfile
			)
		except Exception as e:
			if not isinstance(e, osexception):
				e = osexception(u'Unexpected error', exception=e)
			# Communicate the exception and exit with error
			self.output.put(e)
			sys.exit(1)
		oslogger.info(u'Starting experiment as %s' % self.name)
		# Run the experiment and catch any Exceptions.
		e_run = None
		exp.set_output_channel(self.output)
		try:
			exp.run()
			oslogger.info('experiment finished!')
		except Exception as e:
			if not isinstance(e, osexception):
				e_run = osexception(u'Unexpected error', exception=e)
			else:
				e_run = e
		exp.transmit_workspace()
		# End the experiment and catch any Exceptions. These exceptions are just
		# printed out and not explicitly passed on to the user, because they are
		# less important than the run-related exceptions.
		try:
			exp.end()
		except Exception as e_exp:
			oslogger.error(u'An Exception occurred during exp.end(): %s' % e_exp)
		# Communicate the exception and exit with error
		if e_run is not None:
			self.output.put(e_run)
			sys.exit(1)
		# Exit with success
		sys.exit(0)

	def kill(self):

		oslogger.info(u'killing experiment process')
		try:
			os.kill(
				self.pid,
				signal.SIGKILL if hasattr(signal, u'SIGKILL') else signal.SIGTERM
			)
		except OSError:
			oslogger.warning(u'failed to kill experiment process')
		self.killed = True



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.synth import synth as synth_runtime
from libqtopensesame.items.qtplugin import qtplugin
from libqtopensesame.validators import duration_validator
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'synth', category=u'item')

class synth(synth_runtime, qtplugin):

	"""
	desc:
		GUI controls for the synth item.
	"""

	description = _(u'A basic sound synthesizer')
	help_url = u'manual/stimuli/sound'
	lazy_init = True

	def __init__(self, name, experiment, string=None):

		"""See item."""

		synth_runtime.__init__(self, name, experiment, string)
		qtplugin.__init__(self)

	def init_edit_widget(self):

		"""See qtitem."""

		qtplugin.init_edit_widget(self)
		self.add_combobox_control(u'osc', _(u'Waveform'),
			[u'sine', u'saw', u'square', u'white_noise'])
		self.add_line_edit_control(u'freq',
			_(u'Frequency'),
			info=_(u'In Hertz or as note, e.g. "A1"'))
		self.add_spinbox_control(u'attack', _(u'Attack'),
			min_val=0, max_val=10000000, suffix=_(u' ms'))
		self.add_spinbox_control(u'decay', _(u'Decay'),
			min_val=0, max_val=10000000, suffix=_(u' ms'))
		self.add_doublespinbox_control(u'volume', _(u'Volume'),
			min_val=0, max_val=1, suffix=_(u' x maximum'))
		self.add_line_edit_control(u'pan', _(u'Panning'),
			info=_(u'Positive values toward the right; "left" or "right" for full panning'))
		self.add_spinbox_control(u'length', _(u'Length'),
			min_val=0, max_val=10000000, suffix=_(u' ms'))
		self.add_line_edit_control(u'duration', _(u'Duration'),
			info=_(u'In milliseconds, "sound", "keypress", or "mouseclick"'),
			validator=duration_validator(self, default=u'sound'))



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

from libopensesame.item import item
from libqtopensesame.items.qtautoplugin import qtautoplugin

class quest_staircase_next(item):

	"""
	desc:
		Process a response and adjust staircase level.
	"""

	description = u'Updates the Quest test value based on a response'

	def reset(self):

		"""
		desc:
			Initialize default variables.
		"""

		self.var.response_var = u'correct'

	def run(self):

		"""
		desc:
			Run phase for plug-in.
		"""

		resp = self.var.get(self.var.response_var)
		try:
			resp = int(resp)
		except:
			# Don't process non-float responses
			return
		self.experiment.quest.update(self.var.quest_test_value, resp)
		self.experiment.quest_set_next_test_value()

class qtquest_staircase_next(quest_staircase_next, qtautoplugin):

	"""
	desc:
		The GUI part of the plug-in. Controls are defined in info.json.
	"""

	def __init__(self, name, experiment, script=None):

		"""
		desc:
			Constructor.

		arguments:
			name:		The name of the plug-in.
			experiment:	The experiment object.

		keywords:
			script:		A definition script.
		"""

		quest_staircase_next.__init__(self, name, experiment, script)
		qtautoplugin.__init__(self, __file__)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import re
from pyqode.core import api
from pyqode.qt import QtWidgets
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'pyqode_manager', category=u'extension')


class ConvertIndentationMode(api.Mode):

	"""Comments/uncomments a set of lines using Ctrl+/.
	"""

	def __init__(self):

		super(ConvertIndentationMode, self).__init__()
		self._action_tabs_to_spaces = QtWidgets.QAction(
			_('Convert tabs to spaces'),
			self.editor
		)
		self._action_spaces_to_tabs = QtWidgets.QAction(
			_('Convert spaces to tabs'),
			self.editor
		)

	def on_state_changed(self, state):

		"""
		Called when the mode is activated/deactivated
		"""

		if state:
			self._action_spaces_to_tabs.triggered.connect(self._spaces_to_tabs)
			self.editor.add_action(
				self._action_spaces_to_tabs,
				sub_menu='Python'
			)
			self._action_tabs_to_spaces.triggered.connect(self._tabs_to_spaces)
			self.editor.add_action(
				self._action_tabs_to_spaces,
				sub_menu='Python'
			)
		else:
			self.editor.remove_action(
				self._action_spaces_to_tabs,
				sub_menu='Python'
			)
			self._action_spaces_to_tabs.triggered.disconnect(
				self._spaces_to_tabs
			)
			self.editor.remove_action(
				self._action_tabs_to_spaces,
				sub_menu='Python'
			)
			self._action_tabs_to_spaces.triggered.disconnect(
				self._tabs_to_spaces
			)

	def _spaces_to_tabs(self):

		code = self.editor.toPlainText()
		while True:
			match = re.search(
				u'((?<=\n){indent}+)|(\A{indent}+)'.format(
					indent=u' ' * cfg.pyqode_tab_length
				),
				code
			)
			if not match:
				break
			space_indent = u'\t' * (len(match.group()) // cfg.pyqode_tab_length)
			code = code[:match.start()] + space_indent + code[match.end():]
		self.editor.setPlainText(code)
		self.editor.document().setModified(True)

	def _tabs_to_spaces(self):

		code = self.editor.toPlainText()
		while True:
			match = re.search(u'((?<=\n)\t+)|(\A\t+)', code)
			if not match:
				break
			space_indent = u' ' * len(match.group()) * cfg.pyqode_tab_length
			code = code[:match.start()] + space_indent + code[match.end():]
		self.editor.setPlainText(code)
		self.editor.document().setModified(True)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtWidgets
from libopensesame import plugins
from libopensesame.exceptions import osexception
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.base_subcomponent import base_subcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'extension_manager', category=u'core')


def suspend_events(fnc):

	"""
	desc:
		A decorator that causes all the extension manager to be suspended while
		the function is executed.
	"""

	def inner(self, *args, **kwdict):

		self.extension_manager.suspend()
		retval = fnc(self, *args, **kwdict)
		self.extension_manager.resume()
		return retval

	return inner


class extension_manager(base_subcomponent):

	"""
	desc:
		Initializes GUI extensions, and distributes signals (fires) to the
		relevant extensions.
	"""

	def __init__(self, main_window):

		"""
		desc:
			Constructor.

		arguments:
			main_window:    The main-window object.
		"""

		self.setup(main_window)
		self.main_window.set_busy()
		QtWidgets.QApplication.processEvents()
		self._extensions = []
		self.events = {}
		self.provides = {}
		self._suspended = False
		self._suspended_until = None
		extension_filter = lambda ext_name: False
		for ext_name in plugins.list_plugins(
			_type=u'extensions',
			mode=self.main_window.mode
		):
			if extension_filter(ext_name):
				oslogger.debug(u'filtering extension {}'.format(ext_name))
				continue
			try:
				if self.main_window.options.profile:
					from datamatrix import functional as fnc
					with fnc.profile('profile-loadext-{}.log'.format(ext_name)):
						ext = plugins.load_extension(ext_name, self.main_window)
				else:
					ext = plugins.load_extension(ext_name, self.main_window)
			except Exception as e:
				if not isinstance(e, osexception):
					e = osexception(msg=u'Extension error', exception=e)
				self.notify(
					u'Failed to load extension %s (see debug window for stack trace)'
					% ext_name
				)
				self.console.write(e)
			else:
				self._extensions.append(ext)
				self.register_extension(ext, ext_name)
				if ext.extension_filter is not None:
					extension_filter = ext.extension_filter
		self.main_window.set_busy(False)
		
	def register_extension(self, ext, ext_name=None):
		
		if ext_name is None:
			ext_name = str(ext)
		oslogger.debug(u'installing extension {}'.format(ext_name))
		for event in ext.supported_events():
			if event not in self.events:
				self.events[event] = []
			self.events[event].append(ext)
		for provide in ext.supported_provides():
			if provide in self.provides:
				oslogger.warning(
					u'multiple extensions provide {}'.format(provide)
				)
				continue
			self.provides[provide] = ext

	def __getitem__(self, extension_name):

		"""
		desc:
			Emulates a dict interface for retrieving extensions.

		arguments:
			extension_name:
				desc:    The extension name.
				type:    str

		returns:
			type:    base_extension
		"""

		for ext in self._extensions:
			if ext.name() == extension_name:
				return ext
		raise osexception(u'Extension %s does not exist' % extension_name)

	def __contains__(self, extension_name):

		"""
		arguments:
			extension_name:
				desc:    The extension name.
				type:    str

		returns:
			True if the extension exists, False otherwise.
		"""

		return any(ext.name() == extension_name for ext in self._extensions)

	def fire(self, event, **kwdict):

		"""
		desc:
			Fires an event to all extensions that support the event.

		arguments:
			event:        The event name.

		keyword-dict:
		kwdict:        A dictionary with keywords that are applicable to a
					particular event.
		"""

		if self._suspended_until == event:
			self._suspended = False
			self._suspended_until = None
		if self._suspended:
			oslogger.debug('event suspended {}'.format(event))
			return
		oslogger.debug('firing {}'.format(event))
		if event == u'open_experiment':
			for ext in self._extensions:
				ext.register_ui_files()
		for ext in self.events.get(event, []):
			try:
				ext.fire(event, **kwdict)
			except Exception as e:
				if not isinstance(e, osexception):
					e = osexception(msg=u'Extension error', exception=e)
				self.notify(
					u'Extension %s misbehaved on event %s (see debug window for stack trace)'
					% (ext.name(), event)
				)
				self.console.write(e)

	def provide(self, provide, **kwdict):

		if provide not in self.provides:
			oslogger.debug('cannot not provide {}'.format(provide))
			return None
		ext = self.provides[provide]
		oslogger.debug('{} provides {}'.format(ext, provide))
		try:
			return ext.provide(provide, **kwdict)
		except Exception as e:
			if not isinstance(e, osexception):
				e = osexception(msg=u'Extension error', exception=e)
			self.notify(
				u'Extension %s misbehaved on providing %s (see debug window for stack trace)'
				% (ext.name(), provide)
			)
			self.console.write(e)

	def activate(self, ext_name):

		try:
			ext = self[ext_name]
		except Exception:
			return
		ext.activate()

	def suspend(self):

		"""
		desc:
			Suspends all events.
		"""

		self._suspended = True

	def suspend_until(self, event):

		"""
		desc:
			Suspends all events until a specific event is fired. This is useful
			for situations where you want to supress all events between a
			starting and ending event.

		argument:
			desc:    The unlocking event.
			type:    str
		"""

		self._suspended = True
		self._suspended_until = event

	def resume(self):

		"""
		desc:
			Resumes all events.
		"""

		self._suspended = False



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import pygments.styles
from libqtopensesame.widgets.base_preferences_widget import (
	BasePreferencesWidget
)
from libqtopensesame.misc.config import cfg


class PyQodePreferences(BasePreferencesWidget):
	
	def __init__(self, main_window):

		super(PyQodePreferences, self).__init__(
			main_window,
			ui=u'extensions.pyqode_manager.preferences'
		)
		
	def _before_init_widgets(self):
		
		for style in pygments.styles.get_all_styles():
			self.ui.cfg_pyqode_color_scheme.addItem(style)
		self.ui.cfg_pyqode_fixed_width_nchar.setEnabled(
			cfg.pyqode_fixed_width
		)
		self.ui.label_fixed_width_nchar.setEnabled(
			cfg.pyqode_fixed_width
		)
		self.ui.cfg_pyqode_pep8_ignore.setEnabled(
			cfg.pyqode_pep8_validation
		)
		self.ui.label_pep8_ignore.setEnabled(
			cfg.pyqode_pep8_validation
		)
		self.ui.cfg_pyqode_code_completion_excluded_mimetypes.setEnabled(
			cfg.pyqode_code_completion
		)
		self.ui.label_pyqode_code_completion_excluded_mimetypes.setEnabled(
			cfg.pyqode_code_completion
		)
		self.ui.cfg_pyqode_indentation.currentTextChanged.connect(
			self._toggle_indentation
		)
		self.ui.cfg_pyqode_autopep8_aggressive.setEnabled(
			cfg.pyqode_autopep8
		)
		self._toggle_indentation(cfg.pyqode_indentation)
		
	def _toggle_indentation(self, indentation):
		
		self.ui.cfg_pyqode_tab_length.setEnabled(indentation == 'spaces')
		self.ui.label_tab_length.setEnabled(indentation == 'spaces')



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import time
from libopensesame.oslogging import oslogger
from libqtopensesame.misc import drag_and_drop
from libqtopensesame.misc.base_component import base_component


class base_subcomponent(base_component):

	"""
	desc:
		A base class for all components that require an experiment and theme
		property.
	"""
	
	def name(self):

		"""
		returns:
			desc:    The name of the extension, i.e. the extension class name.
			type:    unicode
		"""

		return safe_decode(self.__class__.__name__, enc=u'utf-8')	
	
	def fire(self, event, **kwdict):

		"""
		desc:
			Calls an event function, if it exists.

		arguments:
			event:
				desc:    The event name, which is handled by a function called
						`event_[event name]`.
				type:    [str, unicode]

		keyword-dict:
			kwdict:        A keyword dictionary with event-specific keywords.
		"""

		if hasattr(self, u'event_%s' % event):
			getattr(self, u'event_%s' % event)(**kwdict)
			
	def _fire_and_time(self, event, **kwdict):
		
		"""
		desc:
			Similar to fire() but also keeps track of the duration of each
			event for benchmarking purposes. This function is used if the
			--performance-profile argument is passed.
		"""
		
		if hasattr(self, u'event_%s' % event):
			t0 = time.time()
			getattr(self, u'event_%s' % event)(**kwdict)
			if event not in self._event_durations:
				self._event_durations[event] = []
			self._event_durations[event].append(time.time() - t0)

	def provide(self, provide, **kwdict):

		"""
		desc:
			Executing a providing function, if it exists.

		arguments:
			provide:
				desc:    The provide name, which is handled by a function called
						`provide_[event name]`.
				type:    [str, unicode]

		keyword-dict:
			kwdict:        A keyword dictionary with provide-specific keywords.

		returns:
			The return value of the providing function.
		"""

		if hasattr(self, u'provide_%s' % provide):
			return getattr(self, u'provide_%s' % provide)(**kwdict)	
			
	def supported_events(self):

		"""
		desc:
			Gives the events that are supported by the extension. This is done
			by introspecting which `event_[event name]` functions exist.

		returns:
			desc:    A list of supported events.
			type:    list
		"""

		events = []
		for event in dir(self):
			if event.startswith(u'event_'):
				events.append(event[6:])
				oslogger.debug(
					u'extension %s supports %s' % (self.name(), event)
				)
		return events

	def supported_provides(self):

		"""
		desc:
			Gives the provides that are supported by the extension. This is
			done by introspecting which `provide_[provide name]` functions
			exist.

		returns:
			desc:    A list of supported provides.
			type:    list
		"""

		provides = []
		for provide in dir(self):
			if provide.startswith(u'provide_'):
				provides.append(provide[8:])
				oslogger.debug(
					u'extension %s provides %s' % (self.name(), provide)
				)
		return provides

	# These properties allow for cleaner programming, without dot references.

	@property
	def experiment(self):
		return self.main_window.experiment

	@property
	def tabwidget(self):
		return self.main_window.ui.tabwidget

	@property
	def theme(self):
		return self.main_window.theme

	@property
	def notify(self):
		return self.main_window.experiment.notify

	@property
	def text_input(self):
		return self.main_window.experiment.text_input

	@property
	def item_store(self):
		return self.experiment.items

	@property
	def extensions(self):
		return self.main_window.extension_manager._extensions

	@property
	def extension_manager(self):
		return self.main_window.extension_manager

	@property
	def overview_area(self):
		return self.main_window.ui.itemtree

	@property
	def console(self):
		return self.main_window.console

	@property
	def pool(self):
		return self.main_window.experiment.pool

	@property
	def locale(self):
		return self.main_window._locale

	@property
	def item_toolbar(self):
		return self.main_window.ui.toolbar_items

	@property
	def main_toolbar(self):
		return self.main_window.ui.toolbar_main


# PEP8 alias
BaseSubcomponent = base_subcomponent



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame import plugins


def libjoystick(experiment, **kwargs):

	"""
	desc:
		A factory that returns a back-end specific joystick module.

	arguments:
		experiment:		The experiment object.

	keyword-dict:
		kwargs:			Keyords to be passed on the joystick init.
	"""

	if experiment.var.canvas_backend == u'psycho':
		raise osexception(
			u'The joystick plug-in does not yet support the psycho back-end'
		)
	backend = u'legacy'
	cls = plugins.load_cls(__file__, cls=backend, mod=backend,
		pkg=u'_libjoystick')
	return cls(experiment, **kwargs)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

from math import hypot
import pygame
from openexp._canvas.legacy import Legacy
from libopensesame.exceptions import osexception
try:
	import android
except ImportError:
	android = None

initialized = False
resolution = 1280, 800  # resolution is hardcoded for now


class Droid(Legacy):

	"""
	This is a canvas backend for Android devices. It is identical to the legacy
	backend, except for the display initialization.
	"""

	@staticmethod
	def init_display(experiment):

		if experiment.resolution() != resolution:
			raise osexception(
				(u'The droid back-end requires a resolution of %d x %d. Your '
				u'display will be scaled automatically to fit devices with '
				u'different resolutions.') % resolution
			)
		# Intialize PyGame
		if not pygame.display.get_init():
			pygame.init()
		experiment.window = pygame.display.set_mode(resolution)
		experiment.surface = pygame.display.get_surface()
		# Set the time functions to use pygame
		experiment._time_func = pygame.time.get_ticks
		experiment._sleep_func = pygame.time.delay
		experiment.time = experiment._time_func
		experiment.sleep = experiment._sleep_func
		# Initialze the Android device if necessary
		if android is not None:
			android.init()
			android.map_key(android.KEYCODE_BACK, pygame.K_ESCAPE)
			dpi = android.get_dpi()
		else:
			# A dummy dpi if we are not on Android
			dpi = 96
		# Log the device characteristics
		info = pygame.display.Info()
		diag = hypot(info.current_w, info.current_h) / dpi
		experiment.var.device_resolution_width = info.current_w
		experiment.var.device_resolution_height = info.current_h
		experiment.var.device_dpi = dpi
		experiment.var.device_screen_diag = diag
		experiment.var.device_is_tablet = u'yes' if diag >= 6 else u'no'
		# Start with a splash screen
		splash = pygame.image.load(experiment.resource('android-splash.jpg'))
		x = resolution[0]/2 - splash.get_width()/2
		y = resolution[1]/2 - splash.get_height()/2
		experiment.surface.blit(splash, (x, y))
		for i in range(10):
			pygame.display.flip()
			pygame.time.delay(100)
		if android is not None and android.check_pause():
			android.wait_for_resume()

	@staticmethod
	def close_display(experiment):

		# On Android, we don't quit the display, as this appears to exit the
		# application altogether.
		if android is None:
			pygame.display.quit()


# Non PEP-8 alias for backwards compatibility
droid = Droid



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp import backend
from libopensesame.oslogging import oslogger
import tempfile
import os

# A list of temporary files that should be cleaned up.
temp_files = []


def Canvas(experiment, *arglist, **kwdict):

	"""
	desc:
		A factory that returns a back-end specific canvas object.

	arguments:
		experiment:
			desc:	The experiment object.
			type:	experiment

	argument-list:
		arglist:	See canvas.__init__().

	keyword-dict:
		kwdict:		See canvas.__init__().
	"""

	cls = backend.get_backend_class(experiment, u'canvas')
	return cls(experiment, *arglist, **kwdict)


def init_display(experiment):

	"""
	desc:
		Calls the back-end specific init_display function.

	arguments:
		experiment:		The experiment object.
		type:			experiment
	"""

	cls = backend.get_backend_class(experiment, u'canvas')
	cls.init_display(experiment)


def close_display(experiment):

	"""
	desc:
		Calls the back-end specific close_display function.

	arguments:
		experiment:		The experiment object.
		type:			experiment
	"""

	cls = backend.get_backend_class(experiment, u'canvas')
	cls.close_display(experiment)


def clean_up(verbose=False):

	"""
	desc:
		Cleans up temporary pool folders.

	keywords:
		verbose:
			desc:	Indicates if debugging output should be provided.
			type:	bool
	"""

	global temp_files
	if verbose:
		oslogger.debug(u"cleaning up temporary pool folders")
	for path in temp_files:
		if verbose:
			oslogger.debug(u"removing '%s'" % path)
		try:
			os.remove(path)
		except Exception as e:
			if verbose:
				oslogger.error(u"failed to remove '%s': %s" % (path, e))


def gabor_file(*arglist, **kwdict):

	"""
	desc:
		Creates a temporary file containing a Gabor patch.

	argument-list:
		arglist:	See canvas.gabor() for a description of arguments.

	keyword-dict:
		kwdict:		See canvas.gabor() for a description of keywords.

	returns:
		A path to the image file.
	"""

	from openexp._canvas import canvas as _canvas
	fd, fname = tempfile.mkstemp(suffix=u'.png')
	temp_files.append(fname)
	im = _canvas._gabor(*arglist, **kwdict)
	im.save(fname)
	os.close(fd)
	return fname


def noise_file(*arglist, **kwdict):

	"""
	desc:
		Creates a temporary file containing a noise patch.

	argument-list:
		arglist:	See canvas.noise_path() for a description of arguments.

	keyword-dict:
		kwdict:		See canvas.noise_path() for a description of keywords.

	returns:
		A path to the image file.
	"""

	from openexp._canvas import canvas as _canvas
	fd, fname = tempfile.mkstemp(suffix=u'.png')
	temp_files.append(fname)
	im = _canvas._noise_patch(*arglist, **kwdict)
	im.save(fname)
	os.close(fd)
	return fname


# Non PEP-8 alias for backwards compatibility
canvas = Canvas



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import tarfile
from libopensesame import misc
from libopensesame.osexpfile import osexpbase


class osexpreader(osexpbase):

	"""
	desc:
		A reader for osexp files of any format.
	"""

	def __init__(self, exp, src):

		"""
		desc:
			Constructor. A side effect of calling this constructor is that all
			files in the file pool of src are extracted to the file pool folder
			of the experiment.

		arguments:
			exp:
				desc:	An experiment object.
				type:	experiment
			src:
				desc:	An OpenSesame script, or the name of an OpenSesame
						experiment file.
				type:	str
		"""

		osexpbase.__init__(self, exp)
		self._src = safe_decode(src)
		if self.format == u'script':
			self._read_script()
		elif self.format == u'scriptfile':
			self._read_scriptfile()
		elif self.format == u'targz':
			self._read_tarfile(format=u'r:gz')
		elif self.format == u'tar':
			self._read_tarfile(format=u'r')
		else:
			raise TypeError(u'"%s" is not an OpenSesame experiment')

	@property
	def script(self):

		"""See osexpbase"""

		return self._script

	def _determine_format(self):

		"""See osexpbase"""

		try:
			# Checking whether a path exists can appartently trigger a
			# ValueError when src is too long. Therefore, we use this awkward
			# construction.
			if not os.path.exists(self._src):
				raise ValueError(u'Path doesn\'t exist)')
		except ValueError:
			return u'script'
		try:
			tarfile.open(self._src, u'r:gz')
			return 'targz'
		except tarfile.ReadError:
			pass
		try:
			tarfile.open(self._src, u'r')
			return 'tar'
		except tarfile.ReadError:
			return 'scriptfile'

	def _read_script(self):

		"""
		visible: False

		desc:
			Reads a script (ie. not a file)
		"""

		self._script = self._src
		self._experiment_path = None

	def _read_scriptfile(self):

		"""
		visible: False

		desc:
			Reads a plain-text script file.
		"""

		self._experiment_path = os.path.dirname(self._src)
		with safe_open(self._src, universal_newline_mode) as fd:
			self._script = fd.read()

	def _read_tarfile(self, format):

		"""
		visible: False

		desc:
			Reads a tar or targz archive.

		arguments:
			format:
				desc:	The format to be used for tarfile.open().
				type:	str
		"""

		self._experiment_path = os.path.dirname(self._src)
		tar = tarfile.open(self._src, format)
		for name in tar.getnames():
			# Tar doesn't return unicode in Python 2
			uname = safe_decode(name)
			folder, fname = os.path.split(uname)
			# Ignore everything except the pool folder
			if folder != u'pool':
				continue
			# Filenames are encoded with U+XXXX notation in the archive. This is
			# necessary to deal with absence of good unicode support in .tar.gz.
			fname = self._syntax.from_ascii(fname)
			pool_folder = safe_str(
				self._pool.folder(),
				enc=misc.filesystem_encoding()
			)
			tar.extract(name, pool_folder)
			from_name = os.path.join(self._pool.folder(), uname)
			to_name = os.path.join(self._pool.folder(), fname)
			os.rename(from_name, to_name)
			os.rmdir(os.path.join(self._pool.folder(), folder))
		# Temporarily extract the script file to the file pool, and remove it
		# right away
		script_path = os.path.join(self._pool.folder(), u'script.opensesame')
		tar.extract(u'script.opensesame', self._pool.folder())
		with safe_open(script_path, universal_newline_mode) as fd:
			self._script = fd.read()
		os.remove(script_path)
